from typing import Protocol, runtime_checkable

from ..service_layer.message_bus import MessageBus


@runtime_checkable
class EconomicModel(Protocol):
    def run(self, bus: MessageBus) -> None:
        """Run the economic model."""
        ...



import pandas as pd
import numpy as np
import re
import logging

from pathlib import Path

from pandas import DataFrame

from steelo.domain import Auto, Year
from steelo.config import settings
from steelo.domain.mappings import GEM_COUNTRY_WS_REGION_MAP
from steelo.utilities.global_variables import (
    STEEL_PLANT_GEM_DATA_YEAR,
    PRODUCTION_GEM_DATA_YEARS,
    SIMULATION_START_YEAR,
)

logging.basicConfig(level=logging.INFO)


# ----------------------------------------------------- FUNCTION DEF -----------------------------------------------------
def load_data() -> tuple[DataFrame, DataFrame, DataFrame, DataFrame]:
    """
    Load the steel plant and historical production data (per plant and regional averages) from csv files.
    """
    logging.info("Loading steel plant data from csv file.")
    plant_data = pd.read_csv(settings.steel_plants_csv)
    production_data = pd.read_csv(settings.hist_production_csv, header=1)
    regional_iron_production_data = pd.read_csv(settings.reg_hist_production_iron_csv)
    regional_steel_production_data = pd.read_csv(settings.reg_hist_production_steel_csv)
    return plant_data, production_data, regional_iron_production_data, regional_steel_production_data


def anonimise_plant_data(data: pd.DataFrame) -> pd.DataFrame:
    """
    Anonimise the plant data by dropping all columns which include any information that could be used to identify the plant (except for its lat-lon coordinates).
    Note: Consider some spatial aggregation of the data later on, since it is easy to identify a plant based on its location.
    """
    logging.info("Anonimising plant data.")
    columns_to_drop = [
        col
        for col in data.columns
        if any(
            keyword in col.lower()
            for keyword in [
                "plant name",  # All columns including the plant name (4)
                "owner",  # All columns including the owner (4)
                "address",  # All columns including the address (2)
            ]
        )
    ] + [
        "Parent",
        "Parent Perm ID",
        "Municipality",
        "Subnational unit (province/state)",
        "GEM wiki page",
    ]
    data_anon = data.drop(columns=columns_to_drop)
    logging.debug(f"Removing columns: {columns_to_drop}")
    return data_anon


def split_spatial_coordinates(data: pd.DataFrame) -> pd.DataFrame:
    """
    Check if there are any plants with missing or invalid coordinates and split them into latitude and longitude columns. Manually added
    missing coordinates for plants 'P100000121145' and 'P100000121251' in the csv file using Google Maps.
    """
    logging.info("Extracting latitude and longitude.")

    def parse_coords(coord):
        if isinstance(coord, str) and "," in coord:
            parts = coord.split(",")
            if len(parts) == 2:
                try:
                    lat, lon = float(parts[0]), float(parts[1])
                    return lat, lon
                except ValueError:
                    pass
        return None, None

    latitudes, longitudes = [], []
    invalid_ids = []

    for i, row in data.iterrows():
        lat, lon = parse_coords(row["Coordinates"])
        if lat is None or lon is None:
            invalid_ids.append((row["Plant ID"], row["Coordinates"]))
        latitudes.append(lat)
        longitudes.append(lon)

    if invalid_ids:
        logging.warning(
            "Invalid coordinates found for Plant IDs:\n%s", "\n".join(f"{pid:>8} {coord}" for pid, coord in invalid_ids)
        )

    data = data.copy()
    data["Latitude"] = latitudes
    data["Longitude"] = longitudes
    return data.drop(columns=["Coordinates", "Coordinate accuracy"])


def filter_inactive_plants(data: pd.DataFrame) -> pd.DataFrame:
    """
    Remove inactive plants (retired, idled, mouthballed).
    """
    logging.info("Filtering inactive plants.")
    # Filter based on operating status
    data_active = data[
        data["Capacity operating status"].isin(["announced", "construction", "operating", "operating pre-retirement"])
    ]
    logging.debug(f"There are {len(data_active)} active plants out of a total of {len(data)} plants.")

    # Refine filter by dropping rows with past retired or idled dates
    # Retired date: Date that iron/steel plant was permanently closed, either after being disassembled or after a period of inactivity (after 5 years of “mothballed”)
    # Idled date: Date that iron/steel plant stops iron or steel production, but has not been disassembled or permanently closed
    # TODO Hannah: Consider filtering out at the time of plant age == average retirement age
    current_date = pd.to_datetime("today").date()
    data_active_date = data_active[
        ~((data_active["Retired Date"] <= str(current_date)) | (data_active["Idled Date"] <= str(current_date)))
    ].reset_index(drop=True)
    logging.debug(
        f"There are {len(data_active_date)} plants with non-past retirement nor idled dates out of {len(data_active)} active plants."
    )
    return data_active_date


def split_plants_into_furnace_groups(data: pd.DataFrame) -> pd.DataFrame:
    """
    Split the plants into furnace groups based on their technology (main production equipment).
    """
    logging.info("Splitting plants into furnace groups based on technology and plant.")
    data_furnace = (
        data.assign(**{"Main production equipment": data["Main production equipment"].str.split(", ")})
        .explode("Main production equipment")
        .reset_index(drop=True)
    )
    data_furnace = data_furnace.rename(columns={"Main production equipment": "Technology"})
    data_furnace.drop(
        [
            "Nominal crude steel capacity (ttpa)",
            "Nominal iron capacity (ttpa)",
        ],
        axis=1,
        inplace=True,
    )
    logging.debug(f"Number of furnace groups: {len(data_furnace)}")
    return data_furnace


def set_unique_furnace_id(data: pd.DataFrame) -> pd.DataFrame:
    """
    Create a unique furnace ID by adding a number to the plant ID  if it is repeated and set it as the Dataframe index.
    """
    logging.info("Setting unique furnace ID.")
    counts = data.groupby("Plant ID").cumcount() + 1
    id_append = counts.astype(str).where(counts > 1).apply(lambda x: f"_{x}" if not pd.isna(x) else "")
    data["Furnace ID"] = data["Plant ID"] + id_append
    data.set_index("Furnace ID", inplace=True)
    # TODO: Remove duplicate furnace groups with the same plant ID, technology, and start date but different operating status (e.g., P100000120035_3 and P100000120035_5).
    # The capacity data is sometimes missing in one of the duplicates; need to make sure to pick the one with non-nan data.
    return data


def rename_capacity_cols(data: pd.DataFrame) -> pd.DataFrame:
    """
    Rename the capacity columns to have consistent names across all products and technologies.
    """
    logging.info("Renaming capacity columns for consistency.")
    for product in ["iron", "steel"]:
        data.rename(
            columns={f"Other/unspecified {product} capacity (ttpa)": f"Nominal other {product} capacity (ttpa)"},
            inplace=True,
        )
        data.rename(
            columns={
                "Nominal BF capacity (ttpa)": "Nominal BF iron capacity (ttpa)",
            },
            inplace=True,
        )
        data.rename(
            columns={
                "Nominal DRI capacity (ttpa)": "Nominal DRI iron capacity (ttpa)",
            },
            inplace=True,
        )
    return data


def treat_non_numeric_values(df: pd.DataFrame, cols: list) -> pd.DataFrame:
    """
    Convert non-numeric values ">0" and "unknown" to NaN and numeric strings to float in the specified columns.
    """
    # TODO: Decide on how to handle ">0" in the data. Currently, converted to Nan for simplicity (information loss!).
    for col in cols:
        df.replace({col: [">0", "unknown"]}, np.nan, inplace=True)
        df[col] = df[col].apply(lambda x: float(x) if isinstance(x, str) and x.replace(".", "", 1).isdigit() else x)
    return df


def allocate_unnamed_tech_to_steel_or_iron(df_row: pd.Series) -> pd.Series:
    """
    Allocate unnamed (other or unknown) technology to steel or iron based its nominal capacity.
    """
    if df_row.Technology == "other" or df_row.Technology == "unknown":
        if df_row["Nominal iron capacity (ttpa)"] > 0 and (
            df_row["Nominal steel capacity (ttpa)"] == 0 or np.isnan(df_row["Nominal steel capacity (ttpa)"])
        ):
            df_row["Technology"] = df_row["Technology"] + " iron"
        elif df_row["Nominal steel capacity (ttpa)"] > 0 and (
            df_row["Nominal iron capacity (ttpa)"] == 0 or np.isnan(df_row["Nominal iron capacity (ttpa)"])
        ):
            df_row["Technology"] = df_row["Technology"] + " steel"
        if df_row["Nominal iron capacity (ttpa)"] > 0 and df_row["Nominal steel capacity (ttpa)"] > 0:
            logging.warning(
                f"Both iron and steel capacities are non-zero for plant {df_row['Plant ID']} and technology {df_row['Technology']}."
            )
            # TODO: Decide how to handle these cases. Currently, the technology is kept as general unknown/other. Ideally, the technology
            # should be split into two rows for iron and steel.
    return df_row


def fill_missing_values_with_global_avg_per_tech(df: pd.DataFrame, col: str) -> pd.DataFrame:
    """
    Fill in missing values in the capacity column with the global average per technology.
    """
    for tech in df["Technology"].unique():
        # Get technology specific rows
        tech_rows = df[df["Technology"] == tech]
        # Calculate global capacity average per technology, excluding NaN values
        cap_avg = tech_rows[col].mean()
        # Fill in missing values
        df.loc[df["Technology"] == tech, col] = df.loc[df["Technology"] == tech, col].fillna(cap_avg)
    return df


def assign_capacity_to_furnace_group(data: pd.DataFrame) -> pd.DataFrame:
    """
    Choose the capacity corresponding to the technology and assign it to each furnace group.
    """
    logging.info("Assigning iron and steel production capacity to each furnace group.")
    for product in ["iron", "steel"]:
        data[f"Nominal {product} capacity (ttpa)"] = data.apply(
            lambda row: row.get(f"Nominal {row['Technology']} {product} capacity (ttpa)", 0), axis=1
        )
        # Add exception for the case when the technology is unknown
        data[f"Nominal {product} capacity (ttpa)"] = data.apply(
            lambda row: (
                row.get(f"Nominal other {product} capacity (ttpa)", 0)
                if row["Technology"] == "unknown"
                else row[f"Nominal {product} capacity (ttpa)"]
            ),
            axis=1,
        )
    return data


def extract_furnace_capacity(data: pd.DataFrame) -> pd.DataFrame:
    """
    Extract the technology used for a certain furnace group (same plant, same technology) and assign the according iron and steel production capacity.
    """
    logging.info("Extracting steel production technology for each furnace group.")

    # For each furnace group, set a unique ID and store the iron and steel-making capacity
    data_furnace = split_plants_into_furnace_groups(data)
    data_furnace_uid = set_unique_furnace_id(data_furnace)
    data_furnace_nn = rename_capacity_cols(data_furnace_uid)
    data_furnace_cap = assign_capacity_to_furnace_group(data_furnace_nn)

    # Drop all unnecessary capacity columns and all other info about the production process and equipment
    capacity_cols_to_drop = [
        col
        for col in data_furnace_cap.columns
        if "capacity" in col.lower()
        and col not in ["Nominal steel capacity (ttpa)", "Nominal iron capacity (ttpa)", "Capacity operating status"]
    ]
    other_cols_to_drop = [
        "Main production process",
        "Detailed production equipment",  # Relevant information included in "Technology" column
        "Iron ore source",
        "Met coal source",
    ]  # Assumed to be irrelevant for the model according to Rafal, take inputs as a black box for now
    cols_to_drop = capacity_cols_to_drop + other_cols_to_drop
    data_furnace_filtered = data_furnace_cap.drop(columns=cols_to_drop)
    logging.debug(f"Removing columns: {cols_to_drop}")

    # Correct non-numeric values and aggregate steel and iron nominal capacities since already clear from the Technology column
    iron_steel_capacity_cols = ["Nominal iron capacity (ttpa)", "Nominal steel capacity (ttpa)"]
    furnace_group_data_numeric = treat_non_numeric_values(data_furnace_filtered, iron_steel_capacity_cols)
    furnace_group_data_numeric = furnace_group_data_numeric.apply(
        lambda row: allocate_unnamed_tech_to_steel_or_iron(row), axis=1
    )
    # TODO: There are 6 FGs where the technology is unknown/other and the capacity is non-zero for both steel and iron. Currenty sumed up.
    # Decide how to handle these cases; linked to the to do in allocate_unnamed_tech_to_steel_or_iron().
    furnace_group_data_numeric["Nominal capacity (ttpa)"] = furnace_group_data_numeric[iron_steel_capacity_cols].sum(
        1, min_count=1
    )  # ensure that the sum of two Nans is Nan
    furnace_group_data_numeric.drop(columns=iron_steel_capacity_cols, inplace=True)

    # Fill in missing capacity values with the global average per technology
    furnace_group_data_filled = fill_missing_values_with_global_avg_per_tech(
        furnace_group_data_numeric, "Nominal capacity (ttpa)"
    )
    return furnace_group_data_filled


def extract_items(data: str) -> list:
    """
    Extracts items from a string containing multiple items, clean up, and convert to lower case.
    """
    # Split by multiple delimiters ([",", ";", " and "]) only if not between parentheses and if "and" is between whitespaces
    split_data = re.split(r",(?![^(]*\))|;(?![^(]*\))|\s+and\s+(?![^(]*\))", data)
    # Remove whitespaces and empty strings
    split_data_full = [item.strip() for item in split_data if (item not in ["", " "]) and (item.strip() != "")]
    # Convert to lowercase and remove duplicates
    split_data_clean = list(set([item.lower() for item in split_data_full]))
    return split_data_clean


def make_steel_product_df(data: pd.DataFrame) -> pd.DataFrame:
    """
    Creates a binary DataFrame with the final steel products (e.g., hotrolled) (columns) produced by each steel plant (rows).
    """
    # TODO: Aggregate final steel products per plant into categories
    logging.info("Creating a DataFrame with the steel products produced by each plant.")
    data.fillna({"Steel products": ""}, inplace=True)
    steel_prod_lists = data["Steel products"].apply(extract_items)
    steel_prod_df = steel_prod_lists.apply(pd.Series).set_index(data["Plant ID"])
    steel_prod_dummies = steel_prod_df.stack().str.get_dummies().groupby(level=0).sum()
    steel_prod_dummies_nodup = (
        steel_prod_dummies.T.groupby(steel_prod_dummies.columns.str.replace("s$", "", regex=True)).sum().T
    )  # Remove plural forms
    steel_prod_bindummies = steel_prod_dummies_nodup.map(lambda x: 1 if x > 0 else 0)  # Convert all values to binary
    logging.debug(
        f"Number of plants producing each steel product:\n{steel_prod_bindummies.sum(axis=0).sort_values(ascending=False)}"
    )
    return steel_prod_bindummies


def zip_year_value(g) -> dict:
    return dict(zip(g["year"], g["value"]))


type GroupByResult = pd.DataFrame | pd.Series


def extract_plant_production_history(raw_production_data: pd.DataFrame) -> GroupByResult:
    """
    Extracts the production for both iron and steel by plant and technology for each year.
    Returns a DataFrame with: (1) rows for each plant ID and technology and (2) columns for each product (iron, steel) and year.
    """
    logging.info("Extracting production data for each plant.")
    data = raw_production_data.set_index("Plant ID")

    # Extract production per plant, technology, and year
    # TODO: Solve for case in which the technology is unknown by making a new row and assigning Technology to unknown (e.g., steel production for plant P100000121188 in 2022). Currently, those cases are ignored.
    data = data.stack().reset_index().rename(columns={0: "value"})
    data[["Production type", "year"]] = data["level_1"].str.extract(r"(?i)(.*) production (\d{4})")
    data = data[~data["Production type"].isin(["Crude steel", "Iron"])]
    data.rename(columns={"Production type": "Technology"}, inplace=True)
    data["Technology"] = (
        data["Technology"].str.replace("EAF steel", "EAF").replace("BOF steel", "BOF").replace("OHF steel", "OHF")
    )
    # TODO: Can't we directly aggreagte steel and iron production?
    data["Iron production (ttpa)"] = data.apply(
        lambda row: row["value"] if row["Technology"] in ["BF", "DRI", "OHF"] else np.nan, axis=1
    )
    data["Steel production (ttpa)"] = data.apply(
        lambda row: row["value"] if row["Technology"] in ["BOF", "EAF"] else np.nan, axis=1
    )

    # Put into a clean format
    data.drop(columns=["level_1", "value"], inplace=True)
    data.dropna(
        subset=["Technology", "year", "Iron production (ttpa)", "Steel production (ttpa)"], how="all", inplace=True
    )

    # Correct non-numeric values
    data_numeric = treat_non_numeric_values(data, ["Iron production (ttpa)", "Steel production (ttpa)"])

    # Aggregate steel and iron historical production since already clear from the Technology column
    data_numeric["Production new"] = data_numeric[["Iron production (ttpa)", "Steel production (ttpa)"]].sum(
        axis=1, min_count=1
    )  # ensure that the sum of two Nans is Nan
    data_numeric = (
        data_numeric.groupby(["Plant ID", "Technology", "year"])["Production new"]
        .sum(min_count=1)  # ensure that the sum of two Nans is Nan
        .unstack("year", fill_value=None)
        .reset_index()
    )
    data_numeric.columns.name = ""  # Remove the year column name
    return data_numeric


def calculate_last_renovation_year(start_date: str, avg_renovation_cycle: int = 20) -> Year | float:
    """
    Calculates the year of the last renovation based on the start date of the plant and the average renovation cycle.
    Renovations are assumed to happen at the beginning of each year and months are truncated.
    """

    start_year = pd.to_datetime(start_date, errors="coerce").year
    if (
        start_year <= SIMULATION_START_YEAR
    ):  # To avoid dates in the future getting a last rennovation year that is in the past
        age = SIMULATION_START_YEAR - start_year

        years_since_last_renovation = age % avg_renovation_cycle

        last_renovation_year = int(SIMULATION_START_YEAR - years_since_last_renovation)

        return Year(last_renovation_year)
    else:
        try:
            return Year(int(start_year))
        except ValueError:
            # FIXME this is probably a bug
            return start_year  # start_year is nan


def impute_missing_dates(data: pd.DataFrame, avg_renovation_cycle: int = 20) -> pd.Series:
    """
    Impute missing 'Start date' values based on "Capacity operating status".

    Rules:
      1. For operating plants with missing start dates, assign a start year by subtracting a random offset
         (between 0 and avg_renovation_cycle) from START_YEAR.
      2. For pre-retirement plants:
         a. Convert 'Retired Date' and 'Pre-retirement Announcement Date' to years.
         b. Compute the 90% confidence interval (CI) of the difference between retirement and announcement years.
         c. For rows with a missing retirement year, add a random offset (within the CI) to the announcement year.
         d. If the retirement year is still missing, sample from the available retirement years.
         e. Set the start year as 20 years before the (imputed) retirement year.
      3. Plants under construction or announced are left unchanged if a start date is present.

    Parameters:
      data : DataFrame containing columns: "Start date", "Capacity operating status", "Retired Date", and "Pre-retirement Announcement Date".
      avg_renovation_cycle : Average renovation cycle in years, default=20

    Returns:
          The "Start date" column after imputing missing values.
    """
    df = data.copy()

    # Convert 'Start date' to datetime
    df["Start date"] = pd.to_datetime(df["Start date"], errors="coerce")

    # Select rows with missing start dates
    missing_dates = df[df["Start date"].isna()].copy()
    logging.info("Handling missing 'Start date' values.")

    # --- Impute for operating plants ---
    operating_mask = missing_dates["Capacity operating status"] == "operating"
    operating_missing = missing_dates[operating_mask]
    if not operating_missing.empty:
        # Sample a random offset between 0 and avg_renovation_cycle and subtract from START_YEAR
        random_offsets = np.random.randint(low=0, high=avg_renovation_cycle, size=operating_missing.shape[0])
        imputed_years = SIMULATION_START_YEAR - random_offsets
        df.loc[operating_missing.index, "Start date"] = pd.Series(
            pd.to_datetime(imputed_years.astype(str)), index=operating_missing.index
        )

    # --- Impute for pre-retirement plants ---
    pre_retirement_mask = missing_dates["Capacity operating status"] == "operating pre-retirement"
    pre_retirement_missing = missing_dates[pre_retirement_mask]
    if not pre_retirement_missing.empty:
        # Convert retirement and announcement dates to year values
        retire_year = pd.to_datetime(pre_retirement_missing["Retired Date"], errors="coerce").dt.year
        announce_year = pd.to_datetime(
            pre_retirement_missing["Pre-retirement Announcement Date"], errors="coerce"
        ).dt.year

        # Compute the 90% confidence interval of the difference between retirement and announcement years
        diff_years = (retire_year - announce_year).dropna()
        if not diff_years.empty:
            ci_lower, ci_upper = diff_years.quantile([0.05, 0.95]).round().astype(int).to_list()

            # For rows with missing retirement year, infer from announcement year with a random offset in the CI
            missing_retire = retire_year.isna()
            if missing_retire.any():
                random_offsets = np.random.randint(
                    low=ci_lower, high=ci_upper, size=announce_year[missing_retire].shape[0]
                )
                retire_year = retire_year.copy()
                retire_year.loc[missing_retire] = announce_year[missing_retire] + random_offsets

            # For any remaining missing retirement years, sample from the available retire_year values
            if retire_year.isna().any():
                sampled_years = retire_year.sample(frac=1, weights=pd.notna(retire_year).astype(float), replace=True)
                sampled_years.index = retire_year.index
                retire_year.fillna(sampled_years, inplace=True)

            # Impute 'Start date' as 20 years before the (imputed) retirement year
            imputed_start_year = (retire_year - 20).astype(int)
            df.loc[pre_retirement_missing.index, "Start date"] = pd.to_datetime(imputed_start_year.astype(str))

    # Return the imputed 'Start date' column
    return df["Start date"]


def filter_relevant_dates(data: pd.DataFrame) -> pd.DataFrame:
    """
    Adds the last renovation year to each furnace group data and drops all irrelevant date columns.
    Adds dates where missing, e.g., for the start date.
    operating under the following ruleset:

    ### Rulset
        1. First we set the data for plants with dates, and assume that renovation occured as scheduled every 20 years ( this is what is being done in calculate last rennovation year)
        2. missing operating data will be sampled from current distribution of under 20 year old furnaces (surely more intelligent methods are available)

        ### For announced and construction furnaces
        1. Announced: If Start date is present, we keep it. If not throw it
        2. For construction furnaces we simply take the start date present
        3. If missing start date we also leave it out for now (it seems that they have researched thigns an examples show that something has happened so that those steel plants are not yet operational)
    """
    logging.info(
        "Filling in missing start dates, calculating the last renovation year per furnace group, and removing all irrelevant dates."
    )

    data["Last renovation year"] = data["Start date"].apply(calculate_last_renovation_year).values
    data["Start date"] = impute_missing_dates(data).copy()
    data["Start date"] = pd.to_datetime(data["Start date"].values, errors="coerce").year.astype("Int64")

    # Remove missing rows with missing start dates

    data = data.rename(columns={"Start date": "Start year"})
    # Fill in missing start dates using plant age if available
    data["Start year"] = data.apply(
        lambda row: (
            STEEL_PLANT_GEM_DATA_YEAR - int(row["Plant age (years)"])
            if pd.isna(row["Start year"]) and pd.notna(pd.to_numeric(row["Plant age (years)"], errors="coerce"))
            else row["Start year"]
        ),
        axis=1,
    )
    # Else, with random years between 2000 and 2013 (most likely period since worldwide production doubled during that time period)
    # TODO: @Rafal to refine based on OECD data
    data["Start year"] = (
        data["Start year"].apply(lambda x: x if pd.notna(x) else np.random.randint(2000, 2014)).astype("Int64")
    )
    data["Last renovation year"] = data["Start year"].apply(calculate_last_renovation_year)
    data.drop(
        columns=[
            "Announced date",
            "Construction date",
            "Pre-retirement Announcement Date",
            "Idled Date",
            "Retired Date",  # TODO: I think we might need the retired year?
            "Plant age (years)",
        ],
        inplace=True,
    )
    return data


def read_technology_lcop_data(path: Path = Auto) -> dict:
    """
    Reads technology LCOP (Levelized Cost of Production) data from a CSV file specified in the settings.
    The CSV file is expected to have two columns where the first column contains the keys and the second column contains the values.

    Returns:
        dict: A dictionary where the keys are from the first column and the values are from the second column of the CSV file.
    """
    if not path:
        path = settings.technology_lcop_csv
    data = pd.read_csv(path)
    return dict(zip(data.iloc[:, 0], data.iloc[:, 1]))


def convert_certifications_to_binary(data: pd.DataFrame) -> pd.DataFrame:
    """
    Combine the three certification-related columns into one and convert it into a binary value. 0: no certification, 1: any valid certification among ISO 14001, ISO 50001, and ResponsibleSteel.
    """
    # TODO Check with Rafal whether more detail is needed (e.g., date, type of certification)
    logging.info("Converting certification columns to binary.")
    cf_cols = ["ISO 14001", "ISO 50001", "ResponsibleSteel Certification"]
    data_bin = data.copy()
    for col in cf_cols:
        data_bin[col] = data[col].apply(
            lambda x: 0 if ((x == "unknown") | (x == "expired") | (not isinstance(x, str) and np.isnan(x))) else 1
        )
    data_bin["Certified"] = data_bin[cf_cols].sum(axis=1).apply(lambda x: 1 if x > 0 else 0)
    data_bin.drop(cf_cols, axis=1, inplace=True)
    return data_bin


def calculate_utilisation(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculate the utilisation of each furnace group in each year by dividing the production by the nominal capacity.
    """
    for year in PRODUCTION_GEM_DATA_YEARS:
        df[f"Utilisation {year}"] = df.apply(
            lambda row: (
                float(row[str(year)]) / float(row["Nominal capacity (ttpa)"])
                if row["Nominal capacity (ttpa)"] != 0
                else 0
            ),  # Set utilisation to 0 if capacity is 0
            axis=1,
        )
        df.drop(columns=str(year), inplace=True)
    return df


def split_production_among_furnace_groups(furnace_group_row: pd.Series, production_df: pd.DataFrame) -> pd.Series:
    """
    Split the production of a plant among the furnace groups with the same technology in the same plant proportionally to the furnace group's capacity.
    Make sure that the furnace group is operational in the year of interest, otherwise, set its production to zero.
    """
    furnace_group = furnace_group_row.copy()
    for year in PRODUCTION_GEM_DATA_YEARS:
        start_year = pd.to_numeric(furnace_group["Start year"], errors="coerce")
        if year >= start_year:  # Make sure production is only allocated to years when the plant was operational
            # Calculate the total capacity of the repeated technologies available in the same plant on that year
            sibling_furnace_groups = production_df[
                (production_df["Plant ID"] == furnace_group["Plant ID"])
                & (production_df["Technology"] == furnace_group["Technology"])
            ]
            sibling_furnace_groups_active = sibling_furnace_groups[
                sibling_furnace_groups["Start year"].astype(int) <= year
            ]
            sibling_capacities = sibling_furnace_groups_active["Nominal capacity (ttpa)"]
            total_capacity = np.sum(sibling_capacities)
            # Calculate the production that should be allocated to the repeated technologies
            if total_capacity > 0:
                furnace_group[str(year)] = (
                    furnace_group[str(year)] * furnace_group["Nominal capacity (ttpa)"] / total_capacity
                )
            else:
                furnace_group[str(year)] = 0
        else:
            furnace_group[str(year)] = 0

        # If production is higher than capacity, set production to capacity
        if furnace_group[str(year)] > furnace_group["Nominal capacity (ttpa)"]:
            furnace_group[str(year)] = furnace_group["Nominal capacity (ttpa)"]
    return furnace_group


def calculate_regional_average_utilisation(
    capacity_df: pd.DataFrame, reg_prod_iron: pd.DataFrame, reg_prod_steel: pd.DataFrame, country_ws_region_map: dict
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Calculate the average utilisation of each technology in each country and WS region by dividing the total production by the aggregated nominal capacity of all furnace groups.
    Assumptions: Set utilisation to 0 if capacity is 0. Cap utlisation at 1 manually if larger than 1 (happens only for complete GEM data).
    """
    # Calculate total capacity per technology per country and WS region
    capacity_df["WS Region"] = capacity_df["Country"].map(country_ws_region_map)
    regional_tot_capacity = capacity_df.groupby(["Country", "Technology"])["Nominal capacity (ttpa)"].sum().to_dict()
    regional_tot_capacity.update(
        capacity_df.groupby(["WS Region", "Technology"])["Nominal capacity (ttpa)"].sum().to_dict()
    )
    # Calculate average utilisation per technology per country and WS region
    for year in PRODUCTION_GEM_DATA_YEARS:
        # Set utilisation to 0 if capacity is 0
        reg_prod_iron[f"Utilisation {year}"] = reg_prod_iron.apply(
            lambda row: (
                row[f"{year}"] / regional_tot_capacity.get((row["Country"], row["Technology"]), 1)
                if regional_tot_capacity.get((row["Country"], row["Technology"]), 1) != 0
                else 0
            ),
            axis=1,
        )
        # Cap utlisation at 1
        reg_prod_iron[f"Utilisation {year}"] = reg_prod_iron[f"Utilisation {year}"].clip(upper=1)
    for year in PRODUCTION_GEM_DATA_YEARS:
        reg_prod_steel[f"{year}"] = reg_prod_steel[f"{year}"] * 1000  # Convert from Miot to ttpa
        # Set utilisation to 0 if capacity is 0
        try:
            reg_prod_steel[f"Utilisation {year}"] = reg_prod_steel.apply(
                lambda row: (
                    row[f"{year}"] / regional_tot_capacity.get((row["WS Region"], row["Technology"]), 1)
                    if regional_tot_capacity.get((row["WS Region"], row["Technology"]), 1) != 0
                    else 0
                ),
                axis=1,
            )
            # Cap utlisation at 1
            reg_prod_steel[f"Utilisation {year}"] = reg_prod_steel[f"Utilisation {year}"].clip(upper=1)
        except KeyError:
            # FIXME jochen - just to make the test pass again
            reg_prod_steel[f"Utilisation {year}"] = 0
    return reg_prod_iron, reg_prod_steel


def fill_missing_utlisation_values(
    df_missing_vals: pd.DataFrame,
    reg_avg_iron: pd.DataFrame,
    reg_avg_steel: pd.DataFrame,
) -> pd.DataFrame:
    """
    Fill in missing historical utilisation values with regional (and global, if regional are not available) averages per technology from World Steel and USGSA.
    """
    df_filled = df_missing_vals.copy()
    for year in PRODUCTION_GEM_DATA_YEARS:
        logging.info(f"Filling missing values in {year}")
        # Fill in missing values with the regional average per technology using the new regional data
        for row in df_filled.iterrows():
            if pd.isna(row[1][f"Utilisation {year}"]):
                if row[1]["Technology"] in reg_avg_iron["Technology"].unique():
                    filtered_iron = reg_avg_iron[
                        (reg_avg_iron["Technology"] == row[1]["Technology"])
                        & (reg_avg_iron["Country"] == row[1]["Country"])
                    ]
                    if not filtered_iron.empty:
                        df_filled.at[row[0], f"Utilisation {year}"] = filtered_iron[f"Utilisation {year}"].values[0]
                elif row[1]["Technology"] in reg_avg_steel["Technology"].unique():
                    filtered_steel = reg_avg_steel[
                        (reg_avg_steel["Technology"] == row[1]["Technology"])
                        # FIXME jochen - just to make the test pass again
                        # & (reg_avg_steel["WS Region"] == row[1]["WS Region"])
                    ]
                    if not filtered_steel.empty:
                        df_filled.at[row[0], f"Utilisation {year}"] = filtered_steel[f"Utilisation {year}"].values[0]
        # Fill in remaining missing values with the global average per technology
        df_filled = fill_missing_values_with_global_avg_per_tech(df_filled, f"Utilisation {year}")
    df_filled.drop(columns=["Country", "WS Region", "Region"], inplace=True)
    return df_filled


def check_column_group_capped_at_1(df: pd.DataFrame, cols: list, var_name: str) -> None:
    """
    Check that a group of columns has values which do not exceed 1. Return a warning and the rows for which at least one of the columns is higher than one
    if it is not the case.
    """
    num_rows_higher_1 = (df[cols].max(axis=1) > 1).sum()
    if num_rows_higher_1 > 0:
        logging.warning(f"{var_name} is >1 for {num_rows_higher_1} rows.")
        logging.info(df[df[cols].max(axis=1) > 1])


def add_historical_production_data(
    plant_furnace_data: pd.DataFrame,
    hist_production_data: pd.DataFrame,
    reg_iron_production_data: pd.DataFrame,
    reg_steel_production_data: pd.DataFrame,
) -> pd.DataFrame:
    """
    Add historical production data to the plant and furnace group data.
    """
    fg_prod_data = pd.merge(
        plant_furnace_data.reset_index(),  # Make Furnace ID a column
        hist_production_data,
        on=["Plant ID", "Technology"],
        how="left",  # keep all furnace groups
    ).set_index("Furnace ID")
    fg_prod_data_corrected = fg_prod_data.apply(
        lambda row: split_production_among_furnace_groups(row, fg_prod_data), axis=1
    )
    reg_avg_utilisation_iron, reg_avg_utilisation_steel = calculate_regional_average_utilisation(
        fg_prod_data_corrected,
        reg_iron_production_data,
        reg_steel_production_data,
        GEM_COUNTRY_WS_REGION_MAP,
    )
    fg_utilisation_data = calculate_utilisation(fg_prod_data_corrected)
    fg_utilisation_data_filled = fill_missing_utlisation_values(
        fg_utilisation_data, reg_avg_utilisation_iron, reg_avg_utilisation_steel
    )
    check_column_group_capped_at_1(
        fg_utilisation_data_filled, [f"Utilisation {year}" for year in PRODUCTION_GEM_DATA_YEARS], "Utilisation"
    )
    return fg_utilisation_data_filled


def split_into_plant_and_furnace_group(
    plant_furnace_data: pd.DataFrame,
    # steel_products_data: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Split the data two dataframes (plant and furnace group data) and save them to file.
    """
    logging.info("Splitting the data into plant and furnace group dataframes and saving them to file.")
    # TODO: Add category of steel product
    plant_df_w_duplicates = plant_furnace_data[
        [
            "Plant ID",
            "Latitude",
            "Longitude",
            "Power source",
            "SOE Status",
            "Parent GEM ID",
            "Workforce size",
            "Certified",
        ]
    ].set_index("Plant ID")
    plant_df = plant_df_w_duplicates[~plant_df_w_duplicates.index.duplicated(keep="first")]
    furnace_group_df = plant_furnace_data[
        [
            "Plant ID",
            "Capacity operating status",
            "Start year",
            "Last renovation year",
            "Technology",
            "Nominal capacity (ttpa)",
        ]
        + [f"Utilisation {year}" for year in PRODUCTION_GEM_DATA_YEARS]
    ]
    # TODO @Jochen: Save to json
    return plant_df, furnace_group_df


# ----------------------------------------------------- WRAPPER -----------------------------------------------------
def execute_preprocessing() -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Execute the preprocessing steps in the correct order.
    """
    logging.info("Executing all preprocessing steps.")
    raw_plant_data, raw_production_data, raw_regional_iron_production_data, raw_regional_steel_production_data = (
        load_data()
    )
    hist_production_data = extract_plant_production_history(raw_production_data)
    anon_data = anonimise_plant_data(raw_plant_data)
    anon_data_latlon = split_spatial_coordinates(anon_data)
    anon_data_latlon_active = filter_inactive_plants(anon_data_latlon)
    anon_data_latlon_active_fc = extract_furnace_capacity(anon_data_latlon_active)
    anon_data_latlon_active_fc_cdates = filter_relevant_dates(anon_data_latlon_active_fc)
    anon_data_latlon_active_fc_cdates_bincf = convert_certifications_to_binary(anon_data_latlon_active_fc_cdates)
    assert isinstance(hist_production_data, pd.DataFrame)
    anon_data_latlon_active_fc_cdates_bincf_hist = add_historical_production_data(
        anon_data_latlon_active_fc_cdates_bincf,
        hist_production_data,
        raw_regional_iron_production_data,
        raw_regional_steel_production_data,
    )
    # steel_products_data = make_steel_product_df(anon_data_latlon_active_fc_cdates_bincf_hist)
    return split_into_plant_and_furnace_group(
        anon_data_latlon_active_fc_cdates_bincf_hist,
        # steel_products_data
    )


if __name__ == "__main__":
    execute_preprocessing()



"""
Command line interface for the steelo package.
"""

import sys
import argparse
from pathlib import Path
import logging

from rich.console import Console

from ..adapters.repositories import InMemoryRepository, JsonRepository, Repository
from ..adapters.dataprocessing.excel_reader import read_plants_into_repository, read_demand_centers_into_repository
from ..adapters.dataprocessing.preprocessing.cost_of_x_data_processing import merge_geographical_data
from ..bootstrap import bootstrap
from ..config import settings
from ..domain.models import Plant
from ..service_layer import get_markers, UnitOfWork
from ..simulation import Simulation, SimulationConfig, SimulationRunner
from ..economic_models import PlantsGoOutOfBusinessOverTime


def show_plants_on_map() -> str:
    console = Console()

    # Initialize the argument parser
    parser = argparse.ArgumentParser(description="Show the location of steel plants on a map.")

    # Add the mp3_url positional argument
    default_csv_path = settings.steel_plants_csv
    default_map_path = settings.root / "output" / "steel_plants_map.html"
    parser.add_argument("--use_csv", action="store_true", help="Whether to use csv input or not. Default is False.")
    parser.add_argument(
        "steel_source", type=str, nargs="?", default=str(default_csv_path), help="Path to the steel csv source file."
    )
    default_production_source = settings.hist_production_csv
    parser.add_argument(
        "production_source",
        type=str,
        nargs="?",
        default=str(default_production_source),
        help="Path to the historical production csv source file.",
    )
    default_gravity_distances_source = settings.gravity_distances_csv
    parser.add_argument(
        "gravity_distances",
        type=str,
        nargs="?",
        default=str(default_gravity_distances_source),
        help="Path to the gravity_distances csv source file.",
    )
    default_iron_production_source = settings.reg_hist_production_iron_csv
    parser.add_argument(
        "iron_production",
        type=str,
        nargs="?",
        default=str(default_iron_production_source),
        help="Path to the historical iron production csv source file.",
    )
    default_steel_production_source = settings.reg_hist_production_steel_csv
    parser.add_argument(
        "steel_production",
        type=str,
        nargs="?",
        default=str(default_steel_production_source),
        help="Path to the historical steel production csv source file.",
    )
    parser.add_argument(
        "html_output", type=str, nargs="?", default=str(default_map_path), help="Path to the html map output file."
    )

    # Parse the command-line arguments
    try:
        args = parser.parse_args()
        use_csv = args.use_csv
        if use_csv:
            steel_source = args.steel_source
            settings.steel_plants_csv = Path(steel_source)
            production_source = args.production_source
            settings.hist_production_csv = Path(production_source)
            gravity_distances = args.gravity_distances
            settings.gravity_distances_csv = Path(gravity_distances)
            iron_production = args.iron_production
            settings.reg_hist_production_iron_csv = Path(iron_production)
            steel_production = args.steel_production
            settings.reg_hist_production_steel_csv = Path(steel_production)
        map_path = args.html_output
    except argparse.ArgumentError as e:
        console.print(f"[red]Argument parsing error: {e}[/red]")
        sys.exit(1)
    except Exception as e:
        console.print(f"[red]Unexpected error during argument parsing: {e}[/red]")
        sys.exit(1)

    # Start drawing the map
    def draw_the_map(plants) -> str:
        repository = InMemoryRepository()
        for plant in plants:
            repository.plants.add(plant)
        markers = get_markers(repository)

        import folium  # type: ignore

        m = folium.Map(location=[20, 0], zoom_start=2)
        for kwargs in markers:
            marker = folium.Marker(**kwargs)
            marker.add_to(m)

        m.save(map_path)

        return str(map_path)

    def run_simulation() -> list[Plant]:
        repository: Repository
        if use_csv:
            repository = read_plants_into_repository()
        else:
            repository = JsonRepository(settings.sample_plants_json_path)

        economic_model = PlantsGoOutOfBusinessOverTime()
        uow = UnitOfWork(repository=repository)
        bus = bootstrap(uow=uow)

        simulation_service = Simulation(bus=bus, economic_model=economic_model)
        simulation_service.run_simulation()
        return simulation_service.bus.uow.plants.list()

    try:
        console.print("[blue]Starting drawing the map[/blue]")
        # run the simulation
        plants = run_simulation()

        # draw the map
        map_path = draw_the_map(plants)
        console.print(f"[green]Map written to[/green]: {map_path}")
        return "Map successfully drawn!"
    except Exception as e:
        raise e
        # console.print(f"[red]Error drawing the map: {e}[/red]")
        # sys.exit(1)


def show_cost_of_x_on_map() -> str:
    console = Console()

    # Initialize the argument parser
    parser = argparse.ArgumentParser(description="Show the cost of debt, equity, and capital.")
    choices = [
        "Cost of debt - renewables",
        "Cost of debt - industrial assets",
        "Cost of equity - renewables",
        "Cost of equity - industrial assets",
        "Cost of capital - renewables",
        "Cost of capital - industrial assets",
    ]
    # Add the options for what to visualize
    parser.add_argument(
        "visualization_type",
        type=str,
        choices=choices,
        help="Type of visualization: 'Cost of debt - renewables','Cost of debt - industrial assets', 'Cost of equity - renewables', 'Cost of equity - industrial assets', 'Cost of capital - renewables', 'Cost of capital - industrial assets'.",
    )

    # Add the csv_source and html_output positional arguments
    default_json_path = settings.cost_of_x_csv
    parser.add_argument(
        "json_source", type=str, nargs="?", default=str(default_json_path), help="Path to the json source file."
    )

    # Parse the command-line arguments
    try:
        args = parser.parse_args()
        visualization_type = args.visualization_type
        json_source = args.json_source
        settings.cost_of_x_csv = Path(json_source)
        default_map_path = settings.root / "output" / f"{visualization_type.replace(' ', '_').lower()}_map.jpeg"
        map_path = default_map_path
    except argparse.ArgumentError as e:
        console.print(f"[red]Argument parsing error: {e}[/red]")
        sys.exit(1)
    except Exception as e:
        console.print(f"[red]Unexpected error during argument parsing: {e}[/red]")
        sys.exit(1)

    import geopandas as gpd  # type: ignore
    import pandas as pd
    import matplotlib.pyplot as plt  # type: ignore
    import matplotlib.colors as mcolors  # type: ignore

    # Start drawing the map
    def plot_choropleth_map(geo_df: gpd.GeoDataFrame, value_column: str, cmap: str = "viridis"):
        """
        Plots a choropleth map with optional handling for zero values.

        Parameters:
            geo_df (GeoDataFrame): A GeoDataFrame containing geometry and data.
            value_column (str): The column to use for coloring the map.
            title (str): The title of the map.
            cmap (str): The colormap to use for the choropleth. Default is 'viridis'.
        """
        # Fill missing values and normalize the value range
        geo_df = gpd.GeoDataFrame(geo_df, geometry="geometry")
        norm = mcolors.Normalize(vmin=geo_df[value_column].min(), vmax=geo_df[value_column].max())

        # Create the plot
        fig, ax = plt.subplots(1, 1, figsize=(20, 5))

        # Plot the main choropleth
        geo_df.plot(
            column=value_column,
            ax=ax,
            legend=True,
            cmap=cmap,
            edgecolor="black",
            linewidth=0.5,
            norm=norm,
            missing_kwds={"color": "white"},
        )

        # Add a title and remove axes
        ax.set_title(value_column, fontsize=16)
        ax.axis("off")

        plt.savefig(map_path)
        plt.close(fig)

        return str(map_path)

    try:
        console.print("[blue]Starting drawing the map[/blue]")
        # draw the map based on the visualization type
        data_cost_of_x_df = pd.read_json(json_source)
        geo_cost_of_x_df = merge_geographical_data(data_cost_of_x_df, "Country code")
        map_path = plot_choropleth_map(geo_cost_of_x_df, visualization_type)
        console.print(f"[green]Map written to[/green]: {map_path}")
        return "Map successfully drawn!"
    except Exception as e:
        console.print(f"[red]Error drawing the map: {e}[/red]")
        sys.exit(1)


def recreate_plants_sample_data():
    """
    Recreate the JSON sample plants data from the current CSV file.
    """
    console = Console()
    read_repository = read_plants_into_repository()
    sample_plants_json_path = settings.sample_plants_json_path
    console.print(f"[blue]Writing to[/blue]: {sample_plants_json_path}")
    write_repository = JsonRepository(sample_plants_json_path)
    plants_list = read_repository.plants.list()
    console.print(f"[blue]Number of plants to write[/blue]: {len(plants_list)}")
    write_repository.plants.add_list(plants_list)
    console.print(f"[green]Sample plants data written to[/green]: {sample_plants_json_path}")


def recreate_demand_center_data(repository=None):
    """
    Recreate the JSON sample demand center data from the current CSV file.
    """

    read_repository = read_demand_centers_into_repository(
        demand_excel_path="../data/data_collection_ultimate_steel.xlsx",
        demand_sheet_name="Steel_Demand_Chris Bataille",
        location_csv="../data/countries.csv",
        repository=repository,
    )
    demand_centers_json_path = settings.sample_demand_center_json_path
    write_repository = JsonRepository(demand_centers_json_path)
    write_repository.demand_center.add_list(read_repository.demand_center.list())
    console = Console()
    console.print(f"[green]Sample demand center data written to[/green]: {demand_centers_json_path}")


def run_full_simulation() -> str:
    """
    Run a full simulation from start to end year with the
    configuration provided via command line arguments.
    """
    console = Console()

    # Initialize the argument parser
    parser = argparse.ArgumentParser(description="Run a full steel model simulation.")

    # Add the options for the simulation
    parser.add_argument("--start-year", type=int, default=2025, help="The year to start the simulation (default: 2025)")
    parser.add_argument("--end-year", type=int, default=2027, help="The year to end the simulation (default: 2027)")
    parser.add_argument(
        "--plants-json",
        type=str,
        default=None,
        help=f"Path to the plants JSON file (default: {settings.sample_plants_json_path})",
    )
    parser.add_argument(
        "--output-file",
        type=str,
        default=None,
        help=f"Path for the output JSON file (default: {settings.intermediate_outputs_simulation_run})",
    )
    parser.add_argument(
        "--demand-excel",
        type=str,
        default=None,
        help=f"Path to the demand excel file (default: {settings.demand_center_xlsx})",
    )
    parser.add_argument(
        "--demand-sheet",
        type=str,
        default="Steel_Demand_Chris Bataille",
        help="Sheet name in the demand excel file (default: 'Steel_Demand_Chris Bataille')",
    )
    parser.add_argument(
        "--location-csv",
        type=str,
        default=None,
        help=f"Path to the location CSV file (default: {settings.location_csv})",
    )
    parser.add_argument(
        "--log-level",
        type=str,
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        help="Set the logging level (default: WARNING)",
    )

    # Parse the command-line arguments
    try:
        args = parser.parse_args()

        # Convert string paths to Path objects if provided
        plants_json_path = Path(args.plants_json) if args.plants_json else settings.sample_plants_json_path
        output_file = Path(args.output_file) if args.output_file else settings.intermediate_outputs_simulation_run
        demand_excel = Path(args.demand_excel) if args.demand_excel else settings.demand_center_xlsx
        location_csv = Path(args.location_csv) if args.location_csv else settings.location_csv

        # Map log level string to logging constants
        log_levels = {
            "DEBUG": logging.DEBUG,
            "INFO": logging.INFO,
            "WARNING": logging.WARNING,
            "ERROR": logging.ERROR,
            "CRITICAL": logging.CRITICAL,
        }
        log_level = log_levels[args.log_level]

        # Create the simulation configuration
        config = SimulationConfig(
            plants_json_path=plants_json_path,
            start_year=args.start_year,
            end_year=args.end_year,
            output_file=output_file,
            demand_center_xlsx=demand_excel,
            demand_sheet_name=args.demand_sheet,
            location_csv=location_csv,
            log_level=log_level,
        )

    except argparse.ArgumentError as e:
        console.print(f"[red]Argument parsing error: {e}[/red]")
        sys.exit(1)
    except Exception as e:
        console.print(f"[red]Unexpected error during argument parsing: {e}[/red]")
        sys.exit(1)

    # Run the simulation
    try:
        console.print("[blue]Starting simulation[/blue]")
        runner = SimulationRunner(config)
        runner.run()

        if config.output_file:
            console.print(f"[green]Simulation results written to[/green]: {config.output_file}")

        return "Simulation completed successfully!"
    except Exception as e:
        console.print(f"[red]Error during simulation: {e}[/red]")
        sys.exit(1)



"""
Base settings.

If there's a .env file in the user's home directory, it will be used.
Otherwise, the .env file in the project root will be used. Environment
variables will override values from .env files.
"""

import os
import sys
import shutil
import logging
import sentry_sdk

from pathlib import Path
from typing import Tuple, Type, ClassVar


from pydantic import DirectoryPath, Field, field_validator, HttpUrl, ValidationError

from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    PydanticBaseSettingsSource,
    DotEnvSettingsSource,
)
from rich import print as rprint

from steelo.domain import Auto
from steelo.utilities.global_variables import STEEL_PLANT_GEM_DATA_YEAR, PRODUCTION_GEM_DATA_YEARS

logger = logging.getLogger(__name__)
ROOT_DIR = Path(__file__).resolve(strict=True).parent.parent.parent

sentry_sdk.init(
    dsn="https://efde20e9df4ed1634bd4fba5f998fa53@o4504836894490624.ingest.us.sentry.io/4508448665042945",
    traces_sample_rate=1.0,
)


def get_steelo_home():
    """
    Don't use this function directly. Use settings.root instead. This function is only
    used to be able to set the default value of the steelo home directory and to be
    able to mock it in tests.
    """
    if (steelo_home := os.getenv("STEELO_HOME")) is not None:
        return Path(steelo_home)
    else:
        return Path.home() / ".steelo"


def get_cbc_path() -> Path | None:
    cbc_path = shutil.which("cbc")
    if cbc_path is None:
        logger.warning("cbc not found. Install it with pip install cbc")
        return None
    return Path(cbc_path)


def get_fixtures_dir() -> Path:
    """Get the fixtures directory path."""
    package_dir = Path(__file__).parent
    if (fixtures_dir := package_dir / "fixtures").exists():
        # installed as package (production) -> return ./fixtures
        return fixtures_dir
    else:
        # running from source (development) -> return $PROJECT_ROOT/data/fixtures
        return ROOT_DIR / "data" / "fixtures"


class Settings(BaseSettings):
    """
    Application settings managed by Pydantic BaseSettings.
    """

    project_root: Path = ROOT_DIR
    root: DirectoryPath = Field(default=get_steelo_home(), alias="STEELO_HOME")
    output_dir: DirectoryPath = Field(default=get_steelo_home() / "output")
    steel_plants_csv: Path = Field(
        default=ROOT_DIR / "data" / f"steel_plants_input_data_{STEEL_PLANT_GEM_DATA_YEAR}.csv"
    )
    hist_production_csv: Path = Field(default=ROOT_DIR / "data" / "historical_production_data.csv")
    reg_hist_production_iron_csv: Path = Field(
        default=ROOT_DIR
        / "data"
        / f"iron_production_{PRODUCTION_GEM_DATA_YEARS[0]}to{PRODUCTION_GEM_DATA_YEARS[-1]}.csv"
    )
    reg_hist_production_steel_csv: Path = Field(
        default=ROOT_DIR
        / "data"
        / f"steel_production_{PRODUCTION_GEM_DATA_YEARS[0]}to{PRODUCTION_GEM_DATA_YEARS[-1]}.csv"
    )
    gravity_distances_csv: Path = Field(default=ROOT_DIR / "data" / "gravity_distances.csv")

    # These fields will be set in the __init__ method to use fixtures_dir
    technology_lcop_csv: Path = Auto
    demand_center_xlsx: Path = Auto
    location_csv: Path = Auto
    new_business_cases: Path = Auto
    cost_of_x_csv: Path = Auto

    intermediate_outputs_simulation_run: Path = Field(default=ROOT_DIR / "outputs" / "pam_simulation_run.json")
    geo_boundaries_url: HttpUrl = Field(
        default=HttpUrl("https://raw.githubusercontent.com/datasets/geo-boundaries-world-110m/master/countries.geojson")
    )
    lp_solver_path: Path | None = get_cbc_path()

    # Store fixtures_dir as a class variable
    _fixtures_dir: ClassVar[Path] = get_fixtures_dir()

    def __init__(self, **data):
        # Set defaults for fields that need fixtures_dir before initializing
        if "technology_lcop_csv" not in data:
            data["technology_lcop_csv"] = self._fixtures_dir / "technology_lcop.csv"
        if "demand_center_xlsx" not in data:
            data["demand_center_xlsx"] = self._fixtures_dir / "Data_collection_ultimate_steel.xlsx"
        if "location_csv" not in data:
            data["location_csv"] = self._fixtures_dir / "countries.csv"
        if "new_business_cases" not in data:
            data["new_business_cases"] = self._fixtures_dir / "2024_12_13_New business cases_MVP.xlsx"
        if "cost_of_x_csv" not in data:
            data["cost_of_x_csv"] = self._fixtures_dir / "cost_of_x.json"
        if "gravity_distances_csv" not in data:
            data["gravity_distances_csv"] = self._fixtures_dir / "gravity_distances.csv"

        super().__init__(**data)

    @field_validator("root", "output_dir", mode="before")
    def path_must_exist_or_create(cls, path: DirectoryPath) -> DirectoryPath:
        """Ensure the path exists or create it."""
        path.mkdir(parents=True, exist_ok=True)
        return DirectoryPath(path)

    model_config = SettingsConfigDict(extra="ignore")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> Tuple[PydanticBaseSettingsSource, ...]:
        """Determine the appropriate .env file"""
        home_env = get_steelo_home() / ".env"
        project_env = ROOT_DIR / ".env"
        if home_env.exists():
            env_file = home_env
        else:
            env_file = project_env

        # Create a new DotEnvSettingsSource with the selected env_file
        new_dotenv_settings = DotEnvSettingsSource(
            settings_cls=settings_cls,
            env_file=env_file,
            env_file_encoding="utf-8",
        )

        # Return the sources in the desired order
        return (
            init_settings,
            env_settings,
            new_dotenv_settings,
            file_secret_settings,
        )

    @property
    def sample_plants_json_path(self) -> Path:
        return self._fixtures_dir / "plants.json"

    @property
    def sample_demand_center_json_path(self) -> Path:
        return self._fixtures_dir / "demand_center.json"


try:
    settings = Settings()
except ValidationError as e:
    rprint("validation error: ", e)
    sys.exit(1)



import random
from typing import TYPE_CHECKING

from ..adapters.repositories.in_memory_repository import PlantInMemoryRepository

if TYPE_CHECKING:
    from ..service_layer import MessageBus


class PlantsGoOutOfBusinessOverTime:
    """
    Uses self.fraction_to_keep to determine which plants to keep
    by randomly selecting a subset of the steel plants.
    """

    def __init__(self, *, fraction_to_keep: float = 0.5) -> None:
        self.fraction_to_keep = fraction_to_keep

    def run(self, bus: "MessageBus") -> None:
        steel_plants = bus.uow.plants.list()
        remaining_plants = random.choices(steel_plants, k=int(len(steel_plants) * self.fraction_to_keep))
        bus.uow.plants = PlantInMemoryRepository()
        bus.uow.plants.add_list(remaining_plants)



from pathlib import Path
from datetime import date
from typing import Self, List, Iterable
from pydantic import BaseModel

from .interface import PlantRepository, FurnaceGroupRepository, DemandCenterRepository
from ...domain import Plant, Volumes, Year, Location, Technology, FurnaceGroup, PointInTime, TimeFrame, DemandCenter


class LocationInDb(BaseModel):
    iso3: str | None
    country: str
    region: str
    lat: float
    lon: float


class TechnologyInDb(BaseModel):
    name: str
    product: str
    technology_readiness_level: int | None
    process_emissions: float | None = None
    emissions_factor: float | None
    energy_consumption: float | None
    bill_of_materials: dict[str, dict[str, dict[str, float]]] | None = None


class TimeFrameInDb(BaseModel):
    start: int
    end: int


class PointInTimeInDb(BaseModel):
    current: int
    time_frame: TimeFrameInDb


class FurnaceGroupInDb(BaseModel):
    furnace_group_id: str
    capacity: Volumes
    status: str
    lifetime: PointInTimeInDb
    last_renovation_date: date
    technology: TechnologyInDb
    historical_production: dict[Year, Volumes]
    utilization_rate: float

    @property
    def to_domain(self) -> FurnaceGroup:
        return FurnaceGroup(
            furnace_group_id=self.furnace_group_id,
            capacity=self.capacity,
            status=self.status,
            lifetime=PointInTime(
                current=Year(self.lifetime.current),
                time_frame=TimeFrame(
                    start=Year(self.lifetime.time_frame.start), end=Year(self.lifetime.time_frame.end)
                ),
            ),
            last_renovation_date=self.last_renovation_date,
            technology=Technology(
                name=self.technology.name,
                product=self.technology.product,
                technology_readiness_level=self.technology.technology_readiness_level,
                process_emissions=self.technology.process_emissions,
                emissions_factor=self.technology.emissions_factor,
                energy_consumption=self.technology.energy_consumption,
                bill_of_materials={}
                if self.technology.bill_of_materials is None
                else self.technology.bill_of_materials,
            ),
            historical_production=self.historical_production,
            utilization_rate=self.utilization_rate,
        )

    @classmethod
    def from_domain(cls, furnace_group: FurnaceGroup) -> Self:
        technology_data = furnace_group.technology.__dict__.copy()
        if technology_data["bill_of_materials"] is not None:
            technology_data["bill_of_materials"] = dict(technology_data["bill_of_materials"])
        return cls(
            furnace_group_id=furnace_group.furnace_group_id,
            capacity=furnace_group.capacity,
            status=furnace_group.status,
            lifetime=PointInTimeInDb(
                current=furnace_group.lifetime.current,
                time_frame=TimeFrameInDb(
                    start=furnace_group.lifetime.start,
                    end=furnace_group.lifetime.end,
                ),
            ),
            last_renovation_date=furnace_group.last_renovation_date,
            technology=TechnologyInDb(**technology_data),
            historical_production=furnace_group.historical_production,
            utilization_rate=furnace_group.utilization_rate,
        )


class PlantInDb(BaseModel):
    plant_id: str
    location: LocationInDb
    furnace_groups: list[FurnaceGroupInDb]
    power_source: str
    soe_status: str
    parent_gem_id: str
    workforce_size: int
    certified: bool
    category_steel_product: set[str]

    def __lt__(self, other: Self) -> bool:
        """
        Make PlantInDb sortable by id. This is useful to keep the order of plants in the json file.
        Which is useful for diffing the file when the list of plants is updated.
        """
        return self.plant_id < other.plant_id

    @property
    def to_domain(self) -> Plant:
        location = Location(**self.location.model_dump())
        furnace_groups = []
        for fg in self.furnace_groups:
            fg_data = fg.model_dump()
            fg_data["technology"] = Technology(**fg_data["technology"])
            fg_data["lifetime"] = PointInTime(**fg.lifetime.model_dump())
            furnace_groups.append(FurnaceGroup(**fg_data))
        plant_data = self.model_dump()
        plant_data["location"] = location
        plant_data["furnace_groups"] = furnace_groups
        return Plant(**plant_data)

    @classmethod
    def from_domain(cls, plant: Plant) -> Self:
        plant_data = plant.__dict__
        plant_data["location"] = LocationInDb(**plant_data["location"].__dict__)
        furnace_groups_in_db = []
        for fg in plant_data["furnace_groups"]:
            fg_data = fg.__dict__
            technology_data = fg_data["technology"].__dict__.copy()
            if not technology_data["bill_of_materials"]:
                print(type(technology_data["bill_of_materials"]))
                technology_data["bill_of_materials"] = dict()
            fg_data["technology"] = TechnologyInDb(**technology_data)
            lifetime = fg_data["lifetime"]
            time_frame_in_db = TimeFrameInDb(start=lifetime.start, end=lifetime.end)
            fg_data["lifetime"] = PointInTimeInDb(current=lifetime.current, time_frame=time_frame_in_db)
            furnace_groups_in_db.append(FurnaceGroupInDb(**fg_data))
        plant_data["furnace_groups"] = furnace_groups_in_db
        return cls(**plant_data)


class PlantListInDb(BaseModel):
    """
    Used to make a list of plants serializable to json.
    """

    root: list[PlantInDb]


class PlantJsonRepository:
    """
    Repository for storing plants in a json file. Uses pydantic models for serialization / deserialization.
    All input and output is done using domain models. Domain models are converted to / from db models
    and then validated and dumped to json.
    """

    _all = None

    def __init__(self, path: Path) -> None:
        self.path = path
        self.seen: set[Plant] = set()

    def _fetch_all(self) -> dict[str, PlantInDb]:
        try:
            if not self.path.exists():
                return {}
            with self.path.open("r", encoding="utf-8") as f:
                plants_in_db = PlantListInDb.model_validate_json(f.read())
            return {plant.plant_id: plant for plant in plants_in_db.root}
        except FileNotFoundError:
            return {}

    @property
    def all(self) -> dict[str, PlantInDb]:
        """
        Cache fetching all models from file.
        """
        if self._all is None:
            self._all = self._fetch_all()
        return self._all

    def get(self, plant_id) -> Plant:
        return self.all[plant_id].to_domain

    def get_by_furnace_group_id(self, furnace_group_id: str) -> Plant:
        for plant_in_db in self.all.values():
            for fg in plant_in_db.furnace_groups:
                if fg.furnace_group_id == furnace_group_id:
                    return plant_in_db.to_domain
        raise KeyError(f"No plant with furnace group id {furnace_group_id}")

    def list(self) -> list[Plant]:
        return [plant_in_db.to_domain for plant_in_db in self.all.values()]

    def _write_models(self, locked: List[PlantInDb]) -> None:  # use List because of list method
        self.path.parent.mkdir(exist_ok=True)
        locked.sort()  # sort by id to keep the order of plants in the file for easier diffing
        plants_in_db = PlantListInDb(root=locked)
        with self.path.open("w", encoding="utf-8") as f:
            f.write(plants_in_db.model_dump_json(indent=2))

    def add(self, plant: Plant) -> None:
        """Add a single plant to the repository."""
        plant_in_db = PlantInDb.from_domain(plant)
        locked = self._fetch_all()
        locked[plant_in_db.plant_id] = plant_in_db
        self._write_models(list(locked.values()))
        self.seen.add(plant)

    def add_list(self, plants: Iterable[Plant]) -> None:
        """Add a list of plant to the repository."""
        plants_in_db = [PlantInDb.from_domain(plant) for plant in plants]
        locked = self._fetch_all()
        for plant_in_db in plants_in_db:
            locked[plant_in_db.plant_id] = plant_in_db
        self._write_models(list(locked.values()))


class FurnaceGroupJsonRepository:
    """
    Repository for storing furnace groups in a json file. Uses pydantic models for serialization / deserialization.
    All input and output is done using domain models. Domain models are converted to / from db models
    and then validated and dumped to json.
    """

    _all = None

    def __init__(self, path: Path) -> None:
        self.path = path
        self.seen: set[FurnaceGroup] = set()

    def _fetch_all(self) -> dict[str, FurnaceGroupInDb]:
        try:
            if not self.path.exists():
                return {}
            with self.path.open("r", encoding="utf-8") as f:
                furnace_groups_in_db = [FurnaceGroupInDb.model_validate_json(line) for line in f]
            return {fg.furnace_group_id: fg for fg in furnace_groups_in_db}
        except FileNotFoundError:
            return {}

    @property
    def all(self) -> dict[str, FurnaceGroupInDb]:
        """
        Cache fetching all models from file.
        """
        if self._all is None:
            self._all = self._fetch_all()
        return self._all

    def get(self, furnace_group_id) -> FurnaceGroup:
        return self.all[furnace_group_id].to_domain

    def list(self) -> list[FurnaceGroup]:
        return [fg.to_domain for fg in self.all.values()]

    def _write_models(self, locked: List[FurnaceGroupInDb]) -> None:
        self.path.parent.mkdir(exist_ok=True)
        locked.sort(
            key=lambda fg: fg.furnace_group_id
        )  # sort by id to keep the order of furnace groups in the file for easier diffing
        with self.path.open("w", encoding="utf-8") as f:
            for fg in locked:
                f.write(fg.model_dump_json(indent=2) + "\n")

    def add(self, furnace_group: FurnaceGroup) -> None:
        """Add a single furnace group to the repository."""
        fg_in_db = FurnaceGroupInDb.from_domain(furnace_group)
        locked = self._fetch_all()
        locked[fg_in_db.furnace_group_id] = fg_in_db
        self._write_models(list(locked.values()))
        self.seen.add(furnace_group)

    def add_list(self, furnace_groups: Iterable[FurnaceGroup]) -> None:
        """Add a list of furnace groups to the repository."""
        fgs_in_db = [FurnaceGroupInDb.from_domain(fg) for fg in furnace_groups]
        locked = self._fetch_all()
        for fg_in_db in fgs_in_db:
            locked[fg_in_db.furnace_group_id] = fg_in_db
        self._write_models(list(locked.values()))

    def to_json(self) -> str:
        """
        Return the JSON content of the furnace groups.
        """
        furnace_groups_in_db = list(self.all.values())
        return "[\n" + ",\n".join(fg.model_dump_json(indent=2) for fg in furnace_groups_in_db) + "\n]"


class DemandCenterInDb(BaseModel):
    demand_center_id: str
    center_of_gravity: LocationInDb
    demand_by_year: dict[Year, Volumes]

    @property
    def to_domain(self) -> DemandCenter:
        location = Location(**self.center_of_gravity.model_dump())
        return DemandCenter(
            demand_center_id=self.demand_center_id,
            center_of_gravity=location,
            demand_by_year=self.demand_by_year,
        )

    @classmethod
    def from_domain(cls, demand_center: DemandCenter) -> Self:
        location_in_db = LocationInDb(**demand_center.center_of_gravity.__dict__)
        return cls(
            demand_center_id=demand_center.demand_center_id,
            center_of_gravity=location_in_db,
            demand_by_year=demand_center.demand_by_year,
        )


class DemandCenterJsonRepository:
    """
    Repository for storing demand centers in a json file. Uses pydantic models for serialization / deserialization.
    All input and output is done using domain models. Domain models are converted to / from db models
    and then validated and dumped to json.
    """

    _all = None

    def __init__(self, path: Path) -> None:
        self.path = path
        self.seen: set[DemandCenter] = set()

    def _fetch_all(self) -> dict[str, DemandCenterInDb]:
        try:
            if not self.path.exists():
                return {}
            with self.path.open("r", encoding="utf-8") as f:
                demand_centers_in_db = [DemandCenterInDb.model_validate_json(line) for line in f]
            return {dc.demand_center_id: dc for dc in demand_centers_in_db}
        except FileNotFoundError:
            return {}

    @property
    def all(self) -> dict[str, DemandCenterInDb]:
        """
        Cache fetching all models from file.
        """
        if self._all is None:
            self._all = self._fetch_all()
        return self._all

    def get(self, demand_center_id) -> DemandCenter:
        return self.all[demand_center_id].to_domain

    def list(self) -> list[DemandCenter]:
        return [dc.to_domain for dc in self.all.values()]

    def _write_models(self, locked: List[DemandCenterInDb]) -> None:
        self.path.parent.mkdir(exist_ok=True)
        locked.sort(
            key=lambda dc: dc.demand_center_id
        )  # sort by id to keep the order of demand centers in the file for easier diffing
        with self.path.open("w", encoding="utf-8") as f:
            for dc in locked:
                f.write(dc.model_dump_json(indent=2) + "\n")

    def add(self, demand_center: DemandCenter) -> None:
        """Add a single demand center to the repository."""
        dc_in_db = DemandCenterInDb.from_domain(demand_center)
        locked = self._fetch_all()
        locked[dc_in_db.demand_center_id] = dc_in_db
        self._write_models(list(locked.values()))
        self.seen.add(demand_center)

    def add_list(self, demand_centers: Iterable[DemandCenter]) -> None:
        """Add a list of demand centers to the repository."""
        dcs_in_db = [DemandCenterInDb.from_domain(dc) for dc in demand_centers]
        locked = self._fetch_all()
        for dc_in_db in dcs_in_db:
            locked[dc_in_db.demand_center_id] = dc_in_db
        self._write_models(list(locked.values()))

    def to_json(self) -> str:
        """
        Return the JSON content of the demand centers.
        """
        demand_centers_in_db = list(self.all.values())
        return "[\n" + ",\n".join(dc.model_dump_json(indent=2) for dc in demand_centers_in_db) + "\n]"


class JsonRepository:
    plants: PlantRepository
    furnace_groups: FurnaceGroupRepository
    demand_centers: DemandCenterRepository

    def __init__(self, path: Path) -> None:
        self.plants = PlantJsonRepository(path)
        self.furnace_groups = FurnaceGroupJsonRepository(path)
        self.demand_centers = DemandCenterJsonRepository(path)



from ..utilities.utils import merge_dicts_bom


class Bill:
    """Base costs/bills related to non-dynamic input variables"""

    IO_mid = {"Value": 124.0, "Unit": "USD/t"}
    IO_high = {"Value": 130.2, "Unit": "USD/t"}
    Met_coal = {"Value": 328.3, "Unit": "USD/GJ"}
    Natural_gas = {"Value": 13.8, "Unit": "USD/GJ"}
    Scrap = {"Value": 344.5, "Unit": "USD/t"}
    Coal = {"Value": 130.0, "Unit": "USD/t"}
    Coking_coal = {"Value": 130.0, "Unit": "USD/t"}
    PCI = {"Value": 130.0, "Unit": "USD/t"}

    Burnt_lime = {"Value": 50.0, "Unit": "USD/t"}
    Burnt_dolomite = {"Value": 50.0, "Unit": "USD/t"}
    Slag = {"Value": 0.0, "Unit": "USD/t"}

    Electricity = {"Value": 0.235, "Unit": "USD/kWh"}
    Natural_gas = {"Value": 13.8, "Unit": "USD/GJ"}
    Hydrogen = {"Value": 35.211, "Unit": "USD/GJ"}

    @classmethod
    def add_additional_bill(cls, new_variable, variable_name):
        setattr(cls, variable_name, new_variable)

    @classmethod
    def to_dict(cls):
        return {key: value for key, value in cls.__dict__.items() if not key.startswith("__") and not callable(key)}


class DBOM:
    """Class for handling bill of materials generation from the dynamic business cases"""

    from steelo.domain.BOM import bom_data

    data = bom_data.copy()

    @classmethod
    def extract_input_values(cls, business_case, metallic_charge=None, reductant=None, share=1):
        # Convert None to slice(None) for wildcard matching
        query = tuple(
            param if param is not None else slice(None) for param in [business_case, metallic_charge, reductant]
        )

        def matches(key_tuple, query_tuple):
            return all((q == slice(None)) or (k == q) for k, q in zip(key_tuple, query_tuple))

        matched_row = {k[3:5]: v for k, v in cls.data.items() if matches(k, query) and k[3] in ["Materials", "Energy"]}

        return matched_row


def translate_input_variables_to_bom(materials, energy, additional_bills, bill=Bill):
    bill_of_materials = {}
    bill_of_energy = {}
    for key, value in materials.items():
        # print(key, value['Value'])
        if value["Value"] > 0:
            try:
                unit_cost = getattr(bill, key.replace(" ", "_"))
            except AttributeError:
                unit_cost = additional_bills[key.replace(" ", "_")]

            bill_of_materials[key] = {
                "demand": value["Value"],
                "unit_cost": unit_cost,
            }
    for key, value in energy.items():
        if key == "Flexible":
            key = "Natural gas"  # let all flexible energy for now be natural gas... for now
        if value["Value"] > 0:
            try:
                unit_cost = getattr(bill, key.replace(" ", "_"))
            except AttributeError:
                unit_cost = additional_bills[key.replace(" ", "_")]

            bill_of_energy[key] = {
                "demand": value["Value"],
                "unit_cost": unit_cost,
            }
    return {"materials": bill_of_materials, "energy": bill_of_energy}


def derive_single_cost(bom, bill=Bill):
    cost = 0
    for key, entry in bom.items():
        for mat, values in entry.items():
            cost += values["demand"] * values["unit_cost"]["Value"]
    return cost


def derive_unnested_feedstock_bom(architype, dbom=DBOM):  # TODO: fix the share thing
    _ = {}  # This is purely meant act as a placeholder to store the bom/unit

    for metallic_charge, items in architype.items():
        # cost for a given metallic charge respective to the current business case
        _input = dbom.extract_input_values(**items)
        materials = {k[1]: v for k, v in _input.items() if k[0] == "Materials"}
        energy = {k[1]: v for k, v in _input.items() if k[0] == "Energy"}

        bom = translate_input_variables_to_bom(materials, energy, _)
        _[f"{metallic_charge.replace(' ', '_')}_bom"] = bom
        # bill.add_additional_bill(bom, f"{metallic_charge.replace(' ', '_')}_{business_case}_bom")
        c = derive_single_cost(
            bom,
        )
        _[f"{metallic_charge.replace(' ', '_')}"] = {"Value": c, "Unit": "USD/t"}
        # bill.add_additional_bill({"Value": c, "Unit": "USD/t"}, f'{metallic_charge.replace(' ', '_')}_{business_case}')
    return _


def extract_steel(bom):
    bill_of = [v for k, v in bom.items() if k in ["Steel_bom", "Scrap_bom"]]
    # print(bill_of)
    if len(bill_of) > 1:
        bill_of_materials = merge_dicts_bom(bill_of)
        return bill_of_materials
    else:
        return bill_of[0]


class BusinessCaseArchitypes:
    """MVP Business case architypes"""

    steel_bof = {
        "Coke": dict(business_case="prep_coke"),
        "Pellets_mid": dict(business_case="prep_pellet", metallic_charge="IO_mid"),
        "Hot metal": dict(business_case="iron_bf", metallic_charge="Pellets_mid", reductant="Coke+PCI"),
        "Steel": dict(business_case="steel_bof", metallic_charge="Hot metal"),
    }

    steel_eaf = {
        "Scrap": dict(business_case="steel_eaf", metallic_charge="Scrap"),
    }
    dri_eaf = {
        "Pellets_high": dict(business_case="prep_pellet", metallic_charge="IO_high"),
        "DRI_high": dict(business_case="iron_dri", metallic_charge="Pellets_high", reductant="Natural gas"),
        "Steel": dict(business_case="steel_eaf", metallic_charge="DRI_high"),
    }

    # h2_dri_eaf_architype = {
    #     "Pellets_high": dict(business_case="prep_pellet", metallic_charge="IO_high"),
    #     "DRI_high": dict(business_case="iron_dri", metallic_charge="Pellets_high", reductant="Hydrogen"),
    #     "Steel": dict(business_case="steel_eaf", metallic_charge="DRI_high"),
    # }

    @classmethod
    def get_all(cls):
        return {
            "BF-BOF": cls.steel_bof_architype,
            "EAF": cls.steel_eaf_architype,
            "DRI-EAF": cls.ng_dri_eaf_architype,
        }

    @classmethod
    def get_architype(cls, key):
        if key == "BF-BOF":
            return cls.steel_bof
        elif key == "EAF":
            return cls.steel_eaf
        elif key == "DRI-EAF":
            return cls.dri_eaf
        else:
            return None



from dataclasses import dataclass
from typing import Any


class Command:
    """Base class for all commands."""


@dataclass
class CloseFurnaceGroup(Command):
    """Close a furnace group."""

    plant_id: str
    furnace_group_id: str


@dataclass
class RenovateFurnaceGroup(Command):
    """Renovate a furnace group."""

    plant_id: str
    furnace_group_id: str


@dataclass
class ChangeFurnaceGroupTechnology(Command):
    """Change the technology of a furnace group."""

    plant_id: str
    furnace_group_id: str
    technology_name: str


@dataclass
class AddFurnaceGroup(Command):
    """Add a furnace group to a plant."""

    plant_id: str
    furnace_group_id: str
    new_furnace: Any



from .models import Plant, Environment
from steelo.utilities.global_variables import ACTIVE_STATUSES


class DataCollector:
    def __init__(self, world_plants: list[Plant], env: Environment, custom_function=None) -> None:
        self.plants = world_plants
        self.env = env
        self.cost_breakdown: dict[str, dict] = {}
        self.trace_capacity: dict[int, dict[str, float]] = {}
        self.trace_price: dict[int, float] = {}
        # self.trace_cost_curve = {}
        self.trace_production: dict[int, float] = {}
        self.step = 0
        self.emissions: dict[int, float] = {}
        self.trace_utilisation_rate: dict[int, dict[str, float]] = {}
        self.capacity_by_technology_and_PAM_status: dict[int, dict[str, dict[bool, float]]] = {}
        self.capacity_deltas: dict[int, dict[str, float]] = {}

        if custom_function is not None:
            pass

    def collect_cost_breakdrown(self):
        """
        This function will return a cost breakdown of the plant, in terms of principal debt, interest, bill of materials, o&m and other opex
        """
        breakdown = {}
        for plant in self.plants:
            breakdown[plant.plant_id] = plant.report_cost_breakdown()
        return breakdown

    def collect_capacity(self):
        """
        This collect the capacity stored in the system environment
        """
        self.env.update_regional_capacity(self.plants)
        return {"Iron": self.env.regional_iron_capacity, "Steel": self.env.regional_steel_capacity}

    def collect_market_iron_steel_price(self, demand: int = 300):
        """
        This function will return the market price of iron and steel
        """
        return {
            "Iron": self.env.extract_price_from_costcurve(demand),
            "Steel": self.env.extract_price_from_costcurve(demand),
        }

    # def collect_params4steel_cost_curve(self):
    #     """
    #     This function will return the steel cost curve and the current demand.
    #     """
    #     return {
    #         "steel_cost_curve": self.env.steel_cost_curve,
    #         "current_demand": self.env.current_demand,
    #         "plants": self.plants,
    #     }

    def collect_emissions_by_plants(self):
        """
        This function will collect the emissions by plants
        """
        emissions = {}
        for plant in self.plants:
            emissions[plant.plant_id] = plant.aggregate_emissions()
        return emissions

    def collect_utilisation_rates(self):
        """collecting furnace_group utilisisation_rates"""
        return {
            fg.furnace_group_id: fg.utilization_rate
            for plant in self.plants
            for fg in plant.furnace_groups
            if fg.status.lower() in ACTIVE_STATUSES
        }

    def collect_capacity_deltas(self):
        delta_added = [plant.added_capacity for plant in self.plants]
        delta_removed = [plant.removed_capacity for plant in self.plants]
        return {"added": sum(delta_added), "removed": sum(delta_removed)}

    def collect_global_steel_production(self):
        """
        This function will collect the production by each operating steel furnace group and return the global total.
        """
        total_production = {}
        for tech in [
            "EAF",
            "BOF",
        ]:  # TODO: Replace with a reference to _allowed_technologies in the technology class post MVP
            total_production[tech] = 0
            for plant in self.plants:
                for fg in plant.furnace_groups:
                    if (
                        (fg.status.lower() in ACTIVE_STATUSES)
                        and (fg.technology.product.lower() == "steel")
                        and (fg.technology.name == tech)
                    ):
                        total_production[tech] += fg.production
        return total_production

    def collect_capacity_by_technology_and_PAM_status(self):
        """
        This function will collect the capacity by technology and PAM status
        """
        capacity = {}
        for tech in [
            "EAF",
            "BOF",
        ]:
            cap_tech_pre_existing = [
                fg.capacity
                for plant in self.plants
                for fg in plant.furnace_groups
                if fg.technology.name == tech and fg.status.lower() in ACTIVE_STATUSES and not fg.created_by_PAM
            ]
            cap_tech_created = [
                fg.capacity
                for plant in self.plants
                for fg in plant.furnace_groups
                if fg.technology.name == tech and fg.status.lower() in ACTIVE_STATUSES and fg.created_by_PAM
            ]
            capacity[tech] = {"pre_existing": sum(cap_tech_pre_existing), "created": sum(cap_tech_created)}
        return capacity

    def production_trade(self):
        pass

    def collect(self):
        self.cost_breakdown[self.step] = self.collect_cost_breakdrown()
        self.trace_capacity[self.step] = self.collect_capacity()
        self.trace_price[self.step] = self.collect_market_iron_steel_price(self.env.current_demand)
        # self.trace_cost_curve[self.step] = self.collect_params4steel_cost_curve()
        self.trace_production[self.step] = self.collect_global_steel_production()
        self.trace_utilisation_rate[self.step] = self.collect_utilisation_rates()
        self.capacity_by_technology_and_PAM_status[self.step] = self.collect_capacity_by_technology_and_PAM_status()
        self.capacity_deltas[self.step] = self.collect_capacity_deltas()
        self.production_trade()
        # self.emissions[self.step] = self.collect_emissions_by_plants().copy() # TODO: fix error and uncomment

        self.step += 1



import pandas as pd
from IPython.display import display, HTML


def display_scrollable_dataframe(df) -> None:
    """
    Display all DataFrame rows in a scrollable element
    """

    with pd.option_context("display.max_rows", None):
        display(HTML(f'<div style="max-height: 400px; overflow-y: scroll;">{df.to_html()}</div>'))


def display_scrollable_series(s) -> None:
    """
    Display all Series rows in a scrollable element, showing the full content of each cell
    """

    with pd.option_context("display.max_rows", None, "display.max_colwidth", None):
        display(
            HTML(
                f'<div style="max-height: 400px; overflow-y: scroll; text-align: left;">{s.to_frame().to_html()}</div>'
            )
        )


def count_appearances(data: pd.DataFrame, col: str) -> None:
    """
    Prints all unique values in a specified column of a DataFrame and counts the number of appearences.
    """

    for i in data[col].unique():
        if pd.isna(i):
            count = data[col].isna().sum()
        else:
            count = data[col].value_counts(dropna=False)[i]
        print(f"Value: {i}; count: {count}")


def merge_dicts_bom(list_of_dicts):
    """A function to merge a list of dicts"""
    merged = {}

    for d in list_of_dicts:
        for top_key, sub_dict in d.items():
            # If 'top_key' isn't in 'merged', create it
            merged.setdefault(top_key, {})

            for inner_key, inner_val in sub_dict.items():
                # If 'inner_key' isn't in 'merged[top_key]', copy this entire sub-dict
                if inner_key not in merged[top_key]:
                    merged[top_key][inner_key] = inner_val.copy()
                else:
                    # Otherwise, just sum the 'demand'
                    merged[top_key][inner_key]["demand"] += inner_val["demand"]

    return merged


def merge_two_dictionaries(target: dict, source: dict) -> dict:
    """
    Merges two dictionaries by summing values for matching keys.
    If a key in source doesn't exist in target, it's added to target.
    """
    for key, value in source.items():
        target[key] = target.get(key, 0) + value
    return target



from pyproj import CRS


def get_laea_crs(geometry):
    """
    Creates a LAEA CRS centered on the geometry's representative point with WGS84 ellipsoid.
    """
    rep_point = geometry.representative_point()
    lat, lon = rep_point.y, rep_point.x
    laea_proj = f"+proj=laea +lat_0={lat} +lon_0={lon} +x_0=0 +y_0=0 +ellps=WGS84 +units=m +no_defs"
    return CRS.from_proj4(laea_proj)



from network_optimisation.hitchcockproblem import HitchcockProblem
from steelo.adapters.repositories import Repository
from steelo.domain.models import SteelAllocations, Year
from steelo.utilities.global_variables import ACTIVE_STATUSES, CAPACITY_LIMIT
from ..events import SteelAllocationsCalculated
from ...service_layer.message_bus import MessageBus


def set_up_HP_steel_trading(repository: Repository, year: Year, global_steel_price=None) -> HitchcockProblem:
    """
    Set up the Hitchcock Problem for steel trading.

    Args:
        repository (Repository): Repository containing data about plants and demand centers.
        year (Year): The year for which the trading is being set up.
        global_steel_price (float, optional): Global steel price to override plant-specific costs.

    Returns:
        HitchcockProblem: Configured Hitchcock Problem instance.
    """
    sources = {}
    sinks = {}
    allocation_costs = {}
    source_costs = {}
    transport_cost_per_tonne_and_km = 0.036
    for demand_center in repository.demand_centers.list():
        sinks[demand_center.demand_center_id] = demand_center.demand_by_year[year]
    for plant in repository.plants.list():
        for furnace_group in plant.furnace_groups:
            if (
                not furnace_group.technology.product.lower() == "steel"
                or furnace_group.status.lower() not in ACTIVE_STATUSES
            ):
                continue
            sources[f"{plant.plant_id}-{furnace_group.furnace_group_id}"] = CAPACITY_LIMIT * furnace_group.capacity
            source_costs[f"{plant.plant_id}-{furnace_group.furnace_group_id}"] = furnace_group.unit_production_cost
            ac_p = {}
            for demand_center in repository.demand_centers.list():
                ac_p[demand_center.demand_center_id] = transport_cost_per_tonne_and_km * plant.distance_to(
                    demand_center.center_of_gravity
                )
            allocation_costs[f"{plant.plant_id}-{furnace_group.furnace_group_id}"] = ac_p
    hp = HitchcockProblem(sources=sources, sinks=sinks, allocation_costs=allocation_costs, source_costs=source_costs)
    return hp


def allocation_from_flows(repository: Repository, mincostFlow: dict) -> SteelAllocations:
    """
    Convert flow data from the Hitchcock Problem solution into steel allocations.

    Args:
        repository (Repository): Repository containing data about plants and demand centers.
        mincostFlow (dict): The flow data from the Hitchcock Problem solution.

    Returns:
        SteelAllocations: Allocated steel flows between plants and demand centers.
    """
    allocations = {}
    for s in mincostFlow:
        for t in mincostFlow[s]:
            if mincostFlow[s][t] > 0:
                try:
                    if "-" not in s:
                        continue
                    plant_id = s.split("-")[0]
                    furnace_group_id = s.split("-")[1]
                    demand_center = repository.demand_centers.get(t)
                    plant = repository.plants.get(plant_id)

                    # Get the actual furnace group object instead of using the ID string
                    furnace_group = next(
                        (fg for fg in plant.furnace_groups if str(fg.furnace_group_id) == furnace_group_id), None
                    )

                    if furnace_group:
                        allocations[(plant, furnace_group, demand_center)] = mincostFlow[s][t]
                except KeyError:
                    continue
    return SteelAllocations(allocations=allocations)


def steel_trade_HP(
    repository: Repository, year: Year, global_steel_price=None, formulation="mincost"
) -> SteelAllocations:
    """
    Solve the steel trade problem using the Hitchcock Problem formulation.

    Args:
        repository (Repository): Repository containing data about plants and demand centers.
        year (Year): The year for which the trading is being set up.
        global_steel_price (float, optional): Global steel price to override plant-specific costs.
        formulation (str, optional): The formulation to use for solving ('lp' or 'mincost'). Defaults to 'lp'.

    Returns:
        SteelAllocations: Allocated steel flows between plants and demand centers.
    """
    hp = set_up_HP_steel_trading(repository, year, global_steel_price)
    if formulation == "lp":
        total_cost, flow = hp.solve_as_lp()
        return allocation_from_flows(repository, flow)
    elif formulation == "mincost":
        min_cost, min_cost_flow = hp.solve_as_min_cost_flow()
        return allocation_from_flows(repository, min_cost_flow)
    raise ValueError("Invalid formulation type. Use 'lp' or 'mincost'.")


def send_allocation_to_bus(allocation: SteelAllocations, bus: MessageBus):
    """
    Send the optimized steel allocations to the message bus.

    Args:
        allocation (SteelAllocations): The optimized steel allocations from steel_trade_HP

    Returns:
        bus (MessageBus): The message bus to send the event to
    """

    if (event := SteelAllocationsCalculated(allocations=allocation)) is not None:
        bus.handle(event)



import pycountry

from geopy import distance  # type: ignore

import reverse_geocoder as rg  # type: ignore


def derive_iso3(lat: float, lon: float, max_distance_km: float = 200) -> str:
    """
    Derive the ISO3 code from latitude and longitude using reverse_geocoder

    Args:
        lat: Latitude (-90 to 90)
        lon: Longitude (-180 to 180)
        max_distance_km: Maximum allowed distance between input and result in km

    Returns:
        ISO3 country code

    Raises:
        ValueError: If coordinates are invalid or result is too far from input
    """
    # Validate coordinate ranges
    if not (-90 <= lat <= 90) or not (-180 <= lon <= 180):
        raise ValueError(
            f"Invalid coordinates: ({lat}, {lon}). Latitude must be between -90 and 90, longitude between -180 and 180."
        )

    try:
        result = rg.search((lat, lon), mode=1)

        if not result or len(result) == 0:
            raise ValueError(f"No result found for coordinates ({lat}, {lon}).")

        result_lat, result_lon = float(result[0]["lat"]), float(result[0]["lon"])
        dist = distance.geodesic((lat, lon), (result_lat, result_lon)).kilometers

        # If the result is too far away, it's likely unreliable
        if dist > max_distance_km:
            raise ValueError(
                f"Result is {dist:.1f}km away from input coordinates. "
                f"This exceeds the maximum allowed distance of {max_distance_km}km. "
                f"Input: ({lat}, {lon}), Result: ({result_lat}, {result_lon})"
            )

        country_code = result[0]["cc"]
        country = pycountry.countries.get(alpha_2=country_code)

        if not country:
            raise ValueError(f"Could not find country with code {country_code} from result.")

        return country.alpha_3

    except Exception as e:
        if isinstance(e, ValueError):
            raise
        raise ValueError(f"Error processing coordinates ({lat}, {lon}): {str(e)}")



from typing import cast

import matplotlib.pyplot as plt
import logging
import cartopy.crs as ccrs  # type: ignore
import cartopy.feature as cfeature  # type: ignore
import geopandas as gpd  # type: ignore
import cartopy.io.shapereader as shpreader  # type: ignore
from cartopy.mpl.geoaxes import GeoAxes  # type: ignore
from folium.map import Figure as FoliumFigure

import pandas as pd
import numpy as np

from matplotlib.figure import Figure

import folium

from ..domain import Volumes, Year
from ..domain.models import SteelAllocations

from ..adapters.repositories.interface import PlantRepository


def plot_cost_curve_per_plant(plants: PlantRepository) -> Figure:
    # Initialize variables for plotting
    current_capacity = 0
    rectangles = []

    # Collect data for rectangle plotting
    for plant in plants.list():
        if hasattr(plant, "average_steel_cost") is False or hasattr(plant, "steel_capacity") is False:
            continue
        width = plant.steel_capacity / 1000000  # Convert to megatonstons
        height = plant.average_steel_cost
        rectangles.append((current_capacity, height, width))
        current_capacity += width

    # Sort rectangles by height (average cost) in ascending order
    rectangles.sort(key=lambda x: x[1])

    # Reset current capacity for sorted rectangles
    current_capacity = 0
    sorted_rectangles = []
    for _, height, width in rectangles:
        sorted_rectangles.append((current_capacity, height, width))
        current_capacity += width

    # Plot the rectangles
    fig, ax = plt.subplots(figsize=(10, 6))
    for x_start, height, width in sorted_rectangles:
        ax.bar(x_start, height, width=width, align="edge", edgecolor="black", color="lightblue")

    # Add labels and grid
    ax.set_xlabel("Cumulative Annual Capacity [megaton per year]", fontsize=12)
    ax.set_ylabel("Cost per Unit [US$/t]", fontsize=12)
    ax.set_title("Cost Curve per Plant", fontsize=14)
    ax.grid(True, linestyle="--", alpha=0.6)

    # Show the plot
    return fig


def plot_steel_cost_curve(curve, demand):
    """
    Plots a steel cost curve (sorted by ascending cost) and indicates the
    marginal producer for a given demand level.

    :param curve: A list of dictionaries of the form
                  [
                    {"cumulative_capacity": float, "production_cost": float},
                    ...
                  ],
                  assumed to be sorted in ascending order of production cost
                  (i.e., a "cost curve").
    :param demand: A float indicating the demanded capacity. The function will
                   highlight the point on the curve where this demand level
                   intersects.
    """

    # Extract x (cumulative capacities) and y (production costs) from the curve
    cumulative_capacities = [point["cumulative_capacity"] for point in curve]
    production_costs = [point["production_cost"] for point in curve]

    # Find maximum capacity to check demand boundaries
    max_capacity = cumulative_capacities[-1] if curve else 0

    # If demand is out of range, we can clip or just note it
    if demand < 0:
        demand = 0
    elif demand > max_capacity:
        print(
            f"Demand ({demand}) exceeds total available capacity ({max_capacity}). "
            "Marginal producer cost will be that of the final producer."
        )
        demand = max_capacity

    # Find the marginal producer:
    # The marginal producer is the first point where cumulative_capacity >= demand
    marginal_producer = None
    for point in curve:
        if point["cumulative_capacity"] >= demand:
            marginal_producer = point
            break

    # Plot the cost curve
    plt.figure(figsize=(8, 5))

    # Often cost curves are visualized as a step plot:
    #   The 'where="post"' makes the step stay at the current y until x changes.
    #   You can also use a simple line plot if you prefer.
    plt.step(cumulative_capacities, production_costs, where="post", label="Cost Curve")

    # Plot a vertical line at the demand level
    plt.axvline(x=demand, color="red", linestyle="--", label="Demand")

    # If we found a valid marginal producer, annotate it
    if marginal_producer:
        marginal_x = demand
        marginal_cost = marginal_producer["production_cost"]

        # A horizontal line at the marginal cost (optional)
        plt.axhline(y=marginal_cost, color="gray", linestyle="--")

        # Mark the intersection
        plt.scatter(marginal_x, marginal_cost, color="red", zorder=5)

        # Add text annotation
        plt.text(
            marginal_x,
            marginal_cost,
            f"  Marginal producer cost = {marginal_cost:.2f}",
            color="red",
            verticalalignment="bottom",
        )

    # Labeling the plot
    plt.title("Steel Cost Curve")
    plt.xlabel("Cumulative Capacity")
    plt.ylabel("Production Cost")
    plt.legend()
    plt.grid(True)

    # Show the plot
    plt.show()


def plot_cost_curve_per_region(plants: PlantRepository) -> Figure:
    # Initialize variables for plotting
    current_capacity = 0
    rectangles = []
    regional_data: dict = {}

    # Collect data for rectangle plotting
    for plant in plants.list():
        if hasattr(plant, "average_steel_cost") is False or hasattr(plant, "steel_capacity") is False:
            continue
        region = plant.location.region
        regional_data[region] = regional_data.get(region, []) + [
            (plant.steel_capacity * plant.average_steel_cost, plant.steel_capacity)
        ]

    for region, data in regional_data.items():
        total_cost = sum([x[0] for x in data])
        total_capacity = sum([x[1] for x in data])
        width = total_capacity / 1000000  # Convert to megatonstons
        height = total_cost / total_capacity
        rectangles.append((current_capacity, height, width))
        current_capacity += total_capacity

    # Reset current capacity for sorted rectangles
    current_capacity = 0
    sorted_rectangles = []
    region_colors = {}
    color_palette = plt.cm.get_cmap("tab20", len(regional_data))

    for idx, (region, data) in enumerate(regional_data.items()):
        total_cost = sum([x[0] for x in data])
        total_capacity = sum([x[1] for x in data])
        width = total_capacity / 1000000  # Convert to megatons
        height = total_cost / total_capacity
        sorted_rectangles.append((current_capacity, height, width, region))
        region_colors[region] = color_palette(idx)
        current_capacity += width

    # Sort rectangles by height (average cost) in ascending order
    sorted_rectangles.sort(key=lambda x: x[1])

    # Sort rectangles by height (average cost) in ascending order
    sorted_rectangles.sort(key=lambda x: x[1])

    # Reset current capacity for sorted rectangles
    current_capacity = 0
    final_rectangles = []
    for x_start, height, width, region in sorted_rectangles:
        final_rectangles.append((current_capacity, height, width, region))
        current_capacity += width

    # Plot the rectangles
    fig, ax = plt.subplots(figsize=(10, 6))
    for x_start, height, width, region in final_rectangles:
        ax.bar(x_start, height, width=width, align="edge", edgecolor="black", color=region_colors[region], label=region)

    # Add legend
    handles, labels = ax.get_legend_handles_labels()
    by_label = dict(zip(labels, handles))
    ax.legend(by_label.values(), by_label.keys(), title="Region")

    # Add labels and grid
    ax.set_xlabel("Cumulative Annual Capacity [megaton per year]", fontsize=12)
    ax.set_ylabel("Cost per Unit [US$/t]", fontsize=12)
    ax.set_title("Cost Curve per Region", fontsize=14)
    ax.grid(True, linestyle="--", alpha=0.6)

    # Show the plot
    return fig


# TODO: Align with Marcus & Shajee on colours and style
def plot_global_heatmap_per_country(
    df: pd.DataFrame,
    var_name: str,
    title: str,
    colormap_type: str = "continous_high_best",
    outlier_treatment: bool = False,
    outlier_side="both",
) -> None:
    """
    Reads in global country level data, converts the country name to the corresponding geometry and plots a global heatmap
    of a certain variable per country. Missing values (and outliers, if applicable) are hatched.
        Inputs:
            - df: DataFrame with the data to plot. Must contain a "Country" column and a column "var_name" with the variable to plot (at least).
            - var_name: Name of the column in df with the variable to plot.
            - title: Title of the plot.
            - colormap_type: Type of colormap to use. Options: "divergent_low_best", "divergent_high_best", "continuous_low_best", "continuous_high_best".
            - outlier_treatment: If True, outliers are set to nan to avoid distorting the color scale.
            - outlier_side: If "both", outliers on both sides are set to nan. If "high"/"low", only high/low outliers are set to nan.
    """

    # Load world geometries from Cartopy's naturalearth_shapefile
    shapefile = shpreader.natural_earth(resolution="110m", category="cultural", name="admin_0_countries")
    reader = shpreader.Reader(shapefile)
    world = reader.records()
    countries_geom = {record.attributes["NAME"]: record.geometry for record in world}
    countries_geom_df = pd.DataFrame(list(countries_geom.items()), columns=["Country", "Geometry"]).set_index("Country")

    # Create a GeoDataFrame
    for i in countries_geom_df.index:
        if i not in df.index:
            logging.warning(f"Warning: {i} not found in df")
    df_with_geom = df.merge(countries_geom_df, left_index=True, right_index=True)
    gdf = gpd.GeoDataFrame(df_with_geom, geometry="Geometry")
    # Rename the column which contains var_name to var_name
    gdf = gdf.rename(columns={col: var_name for col in gdf.columns if str(var_name) in str(col)})

    # Set outliers to nan to avoid distorting the color scale
    if outlier_treatment:
        mean = gdf[var_name].mean()
        if outlier_side in ["both", "high"]:
            gdf.loc[gdf[var_name] > 2 * mean, var_name] = np.nan
        if outlier_side in ["both", "low"]:
            gdf.loc[gdf[var_name] < 0.5 * mean, var_name] = np.nan

    # Set colormap
    if "divergent" in colormap_type:
        colormap = "RdBu"
    elif "continuous" in colormap_type:
        colormap = "Blues"
    if "low_best" in colormap_type:
        colormap = colormap + "_r"

    # Plot heatmap
    fig, ax = plt.subplots(1, 1, figsize=(15, 10), subplot_kw={"projection": ccrs.PlateCarree()})
    assert isinstance(ax, GeoAxes)
    ax.coastlines()
    ax.add_feature(cfeature.BORDERS, linestyle=":")
    ax.add_feature(cfeature.LAND, edgecolor="black")
    gdf.plot(
        ax=ax,
        column=var_name,
        cmap=colormap,
        missing_kwds={
            "color": "none",
            "edgecolor": "black",
            "alpha": 0.5,
            "hatch": "//////",
            "label": "Missing values",
        },
    )
    sm = plt.cm.ScalarMappable(cmap=colormap, norm=plt.Normalize(vmin=gdf[var_name].min(), vmax=gdf[var_name].max()))
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax, orientation="horizontal", pad=0.05, shrink=0.5, label=var_name)
    cbar.ax.tick_params(labelsize=16)
    cbar.set_label(var_name, fontsize=18)
    ax.set_title(title, fontsize=20)
    fig.tight_layout()
    plt.show()
    plt.close()


#######################
#######################
#######################


def plot_allocation_on_map_2(allocation: SteelAllocations, chosen_year: Year):
    """
    allocation.allocations is assumed to be a dictionary keyed by
    (plant, furnace_group, demand_center) -> volume.
    Each 'plant' and 'demand_center' has an 'id' (here used as plant.plant_id,
    demand_center.demand_center_id) and a 'location' with 'lat' and 'lon'.
    The 'furnace_group' has a 'technology' with a 'name'.
    """
    # 1) Create the base map
    m = folium.Map(location=[0, 0], zoom_start=2)

    # 2) Separate arcs by technology (so each tech can be toggled as a separate layer)
    arcs_by_tech: dict[str, list[dict]] = {}
    max_allocation = max(allocation.allocations.values())

    # Predefine known technologies (or create them on the fly)
    known_techs = ["BOF", "EAF", "OHF", "other", "unknown"]
    for tech in known_techs:
        arcs_by_tech[tech] = []

    # 3) Build a GeoJSON Feature for each arc
    for (plant, furnace_group, demand_center), volume in allocation.allocations.items():
        furnace_group = plant.get_furnace_group(furnace_group.furnace_group_id)
        tech_name = furnace_group.technology.name
        if tech_name not in arcs_by_tech:
            arcs_by_tech[tech_name] = []

        normalized_weight = (volume / max_allocation) * 10
        color_map = {
            "BOF": "black",
            "EAF": "darkgreen",
            "OHF": "pink",
            "other": "red",
            "unknown": "red",
        }
        default_color = color_map.get(tech_name, "blue")

        # Build a GeoJSON Feature (LineString)
        # Note we now add "volume" to properties for the tooltip
        feature = {
            "type": "Feature",
            "geometry": {
                "type": "LineString",
                "coordinates": [
                    [plant.location.lon, plant.location.lat],
                    [demand_center.center_of_gravity.lon, demand_center.center_of_gravity.lat],
                ],
            },
            "properties": {
                "start_node_id": plant.plant_id,  # ID of the plant
                "end_node_id": demand_center.demand_center_id,  # ID of the demand center
                "default_color": default_color,
                "default_weight": normalized_weight,
                "tech_name": tech_name,
                "volume": volume,  # <-- Add the volume here
            },
        }
        arcs_by_tech[tech_name].append(feature)

    # 4) Style function for arcs
    def style_function(feature):
        """Style each arc based on its default properties."""
        return {
            "color": feature["properties"]["default_color"],
            "weight": feature["properties"]["default_weight"] * 3,  # Multiply by 6 for a thick base
            "opacity": 1.0,
        }

    # 5) Helper function to add a GeoJson layer and register it
    def add_geojson_layer(map_obj, geojson_features, layer_name):
        """Create a GeoJson layer and register it in window.arcsLayers."""
        if not geojson_features:
            return None
        geojson_data = {"type": "FeatureCollection", "features": geojson_features}

        # Attach a tooltip that displays the "volume" property on hover
        layer = folium.GeoJson(
            data=geojson_data,
            style_function=style_function,
            name=layer_name,
            tooltip=folium.GeoJsonTooltip(fields=["volume"], aliases=["Volume: "]),
        )
        layer.add_to(map_obj)

        # Inject a small <script> to push this layer into window.arcsLayers
        script_str = f"""
        <script>
        (function() {{
            // Once the layer is added, push it into window.arcsLayers
            var gLayer = {layer.get_name()};
            gLayer.on('add', function(e) {{
                console.log("Layer added to arcsLayers:", gLayer);
                if (!window.arcsLayers) {{
                    window.arcsLayers = [];
                }}
                window.arcsLayers.push(gLayer);
            }});
        }})();
        </script>
        """
        layer.add_child(folium.Element(script_str))
        return layer

    # Create one layer per technology
    for tech_name, features in arcs_by_tech.items():
        add_geojson_layer(m, features, tech_name)

    # 6) Add Markers (plants & demand centers)
    #    We'll use folium.Html(..., script=True) + folium.Popup to preserve <button onclick=...>
    marker_fg = folium.FeatureGroup(name="Nodes", show=True)
    marker_fg.add_to(m)

    # Ensure marker_fg is properly referenced in the script_zoom
    script_zoom = f"""
    <script>
    (function() {{
      var markerLayer = {marker_fg.get_name()};
      var myMap = {m.get_name()};  // The main Leaflet map object

      // On zoomend, check the current zoom level.
      myMap.on('zoomend', function() {{
        var currentZoom = myMap.getZoom();
        // If zoom < 7, remove the layer group; else add it back
        if (currentZoom < 7) {{
          myMap.removeLayer(markerLayer);
        }} else {{
          myMap.addLayer(markerLayer);
        }}
      }});

      // Optionally, run it once at startup if you want them hidden initially when zoom < 7
      var initialZoom = myMap.getZoom();
      if (initialZoom < 7) {{
        myMap.removeLayer(markerLayer);
      }}
    }})();
    </script>
    """
    root: FoliumFigure = cast(FoliumFigure, m.get_root())
    root.html.add_child(folium.Element(script_zoom))
    marker_fg.add_to(m)

    for (plant, furnace_group, demand_center), _ in allocation.allocations.items():
        # Example: Summarize capacities by technology
        plant_capacity_by_tech = {}
        for fg in plant.furnace_groups:
            if fg.technology.product != "Steel":  # Skip non-steel technologies
                continue
            if fg.technology.name not in plant_capacity_by_tech:
                plant_capacity_by_tech[fg.technology.name] = fg.capacity
            else:
                plant_capacity_by_tech[fg.technology.name] = Volumes(
                    plant_capacity_by_tech[fg.technology.name] + fg.capacity
                )

        # Build plant popup HTML
        tech_capacity_str = ""
        for tname in plant_capacity_by_tech:
            tech_capacity_str += f"{tname}: {plant_capacity_by_tech[tname]:,.2f} t <br/>"
            tech_capacity_str += f"{tname}: {plant_capacity_by_tech[tname]:,.2f} t <br/>"
        plant_html_str = f"""
        <b>Plant: {plant.plant_id}</b><br/>
        {tech_capacity_str}
        <button onclick="window.toggleNode('{plant.plant_id}')">
            Toggle Node
        </button>
        """
        plant_html = folium.Html(plant_html_str, script=True)
        plant_popup = folium.Popup(plant_html, max_width=250)
        folium.Marker(
            location=[plant.location.lat, plant.location.lon],
            popup=plant_popup,
            icon=folium.Icon(icon="industry", color="blue", prefix="fa"),
        ).add_to(marker_fg)

        # Demand center marker
        demand_html_str = f"""
        <b>Demand Center: {demand_center.demand_center_id}</b><br/>
        Demand: {demand_center.demand_by_year[chosen_year]:,.2f} t</br>
        Demand: {demand_center.demand_by_year[chosen_year]:,.2f} t</br>
        <button onclick="window.toggleNode('{demand_center.demand_center_id}')">
            Toggle Node
        </button>
        """
        demand_html = folium.Html(demand_html_str, script=True)
        demand_popup = folium.Popup(demand_html, max_width=250)
        folium.Marker(
            location=[demand_center.center_of_gravity.lat, demand_center.center_of_gravity.lon],
            popup=demand_popup,
            icon=folium.Icon(icon="cart-shopping", color="red", prefix="fa"),
        ).add_to(marker_fg)

    marker_fg.add_to(m)

    # 7) Inject custom JavaScript for node toggling, highlight, and refresh
    custom_js = """
    <script>
    // Global arrays
    if (!window.selectedNodes) {
        window.selectedNodes = [];
    }
    if (!window.arcsLayers) {
        window.arcsLayers = [];
    }

    window.toggleNode = function(nodeId) {
        console.log("toggleNode called for node:", nodeId);
        var idx = window.selectedNodes.indexOf(nodeId);
        if (idx >= 0) {
            // Node is already selected, remove it
            window.selectedNodes.splice(idx, 1);
        } else {
            // Otherwise, add it
            window.selectedNodes.push(nodeId);
        }
        window.updateArcStyles();
    }

    window.clearSelections = function() {
        window.selectedNodes = [];
        window.updateArcStyles();
    }

    window.updateArcStyles = function() {
        console.log("updateArcStyles called. SelectedNodes:", window.selectedNodes);
        if (!window.arcsLayers) return;

        var noSelection = (window.selectedNodes.length === 0);

        window.arcsLayers.forEach(function(geoJsonLayer) {
            geoJsonLayer.eachLayer(function(layer) {
                var props = layer.feature && layer.feature.properties;
                if (!props) return;

                var startId = props.start_node_id;
                var endId   = props.end_node_id;
                var defColor = props.default_color;
                var defWeight = props.default_weight;

                // Check if the arc is connected to any selected node
                var connected = false;
                for (var i=0; i < window.selectedNodes.length; i++) {
                    var selNode = window.selectedNodes[i];
                    if (startId === selNode || endId === selNode) {
                        connected = true;
                        break;
                    }
                }

                if (noSelection) {
                    // Revert to normal
                    layer.setStyle({
                        color: defColor,
                        weight: defWeight,
                        opacity: 0.6
                    });
                } else if (connected) {
                    // Highlight
                    layer.setStyle({
                        color: defColor,
                        weight: defWeight * 3,
                        opacity: 1.0
                    });
                } else {
                    // Fade
                    layer.setStyle({
                        color: "gray",
                        weight: defWeight,
                        opacity: 0.3
                    });
                }
            });
        });
    }
    </script>
    """
    root.html.add_child(folium.Element(custom_js))

    # 8) Add a “Refresh” button in a fixed position on the map
    refresh_button_html = """
    <div style="
        position: fixed;
        top: 10px;
        left: 50px;
        z-index: 9999;
        background-color: white;
        padding: 5px 10px;
        border: 2px solid grey;
    ">
        <button onclick="window.clearSelections()">Refresh Highlight</button>
    </div>
    """
    root.html.add_child(folium.Element(refresh_button_html))

    script_zoom = f"""
    <script>
    (function() {{
      var markerLayer = {marker_fg.get_name()};
      var myMap = {m.get_name()};  // The main Leaflet map object

      // On zoomend, check the current zoom level.
      myMap.on('zoomend', function() {{
        var currentZoom = myMap.getZoom();
        // If zoom < 7, remove the layer group; else add it back
        if (currentZoom < 7) {{
          myMap.removeLayer(markerLayer);
        }} else {{
          myMap.addLayer(markerLayer);
        }}
      }});

      // Optionally, run it once at startup if you want them hidden initially when zoom < 7
      var initialZoom = myMap.getZoom();
      if (initialZoom < 7) {{
        myMap.removeLayer(markerLayer);
      }}
    }})();
    </script>
    """

    # 5. Add the JS to the final HTML
    root.html.add_child(folium.Element(script_zoom))

    # 9) Add LayerControl so user can toggle each technology layer
    folium.LayerControl().add_to(m)

    # Finally return the map
    return m


##########################################
##########################################


# def build_dash_map_app(allocation, chosen_year: int = 2025):
#     """
#     Build a Dash Leaflet app that replicates (most of) the Folium functionality
#     from `plot_allocation_on_map_2`.

#     :param allocation: An object with .allocations (dict) keyed by
#                        (plant, furnace_group, demand_center) -> volume
#                        and presumably other info about plants/demand centers.
#     :param chosen_year: Which year's demand to display.
#     :return: dash.Dash app instance (not yet run).
#     """

#     # Create the Dash app
#     app = dash.Dash(__name__)

#     # -------------------------------------------------
#     # 1) Build arcs_by_tech
#     # -------------------------------------------------
#     arcs_by_tech = {}
#     known_techs = ["BOF", "EAF", "OHF", "other", "unknown"]

#     if allocation.allocations:
#         max_allocation = max(allocation.allocations.values())
#     else:
#         max_allocation = 1.0

#     # Initialize each known tech
#     for tech in known_techs:
#         arcs_by_tech[tech] = []

#     # Color mapping
#     color_map = {
#         "BOF": "black",
#         "EAF": "darkgreen",
#         "OHF": "pink",
#         "other": "red",
#         "unknown": "red",
#     }

#     # Populate arcs_by_tech
#     for (plant, furnace_group, demand_center), volume in allocation.allocations.items():
#         tech_name = furnace_group.technology.name
#         if tech_name not in arcs_by_tech:
#             arcs_by_tech[tech_name] = []

#         normalized_weight = (volume / max_allocation) * 3  # scale from 0..3
#         default_color = color_map.get(tech_name, "blue")

#         feature = {
#             "type": "Feature",
#             "geometry": {
#                 "type": "LineString",
#                 "coordinates": [
#                     [plant.location.lon, plant.location.lat],
#                     [demand_center.center_of_gravity.lon, demand_center.center_of_gravity.lat],
#                 ],
#             },
#             "properties": {
#                 "start_node_id": plant.plant_id,
#                 "end_node_id": demand_center.demand_center_id,
#                 "default_color": default_color,
#                 "default_weight": normalized_weight,
#                 "tech_name": tech_name,
#                 "volume": volume,
#                 "arrow": True,  # Indicate that this feature should be rendered as an arrow
#             },
#         }
#         arcs_by_tech[tech_name].append(feature)

#     # -------------------------------------------------
#     # 2) Build marker layer for plants and demand centers
#     # -------------------------------------------------
#     marker_features = []

#     def plant_to_feature(plant):
#         # Summarize capacities
#         tech_capacity_str = []
#         for fg in plant.furnace_groups:
#             if fg.technology.product == "Steel":  # skip if not steel
#                 tech_capacity_str.append(f"{fg.technology.name}: {fg.capacity:.2f} Mtpa")
#         popup_str = f"Plant: {plant.plant_id}<br>" + "<br>".join(tech_capacity_str)

#         return {
#             "type": "Feature",
#             "geometry": {"type": "Point", "coordinates": [plant.location.lon, plant.location.lat]},
#             "properties": {
#                 "node_id": plant.plant_id,
#                 "popup": popup_str,
#                 "iconColor": "blue",
#             },
#         }

#     def dc_to_feature(demand_center, year):
#         popup_str = (
#             f"Demand Center: {demand_center.demand_center_id}<br>" f"Demand: {demand_center.demand[year]:,.2f} Mt"
#         )
#         return {
#             "type": "Feature",
#             "geometry": {
#                 "type": "Point",
#                 "coordinates": [demand_center.center_of_gravity.lon, demand_center.center_of_gravity.lat],
#             },
#             "properties": {
#                 "node_id": demand_center.demand_center_id,
#                 "popup": popup_str,
#                 "iconColor": "red",
#             },
#         }

#     # Gather unique (plant, demand_center) from the allocation
#     for (plant, furnace_group, demand_center), _ in allocation.allocations.items():
#         marker_features.append(plant_to_feature(plant))
#         marker_features.append(dc_to_feature(demand_center, chosen_year))

#     marker_collection = {"type": "FeatureCollection", "features": marker_features}

#     # -------------------------------------------------
#     # 3) Define style functions
#     # -------------------------------------------------
#     def style_arcs(feature):
#         """Default style for arcs (when no highlight)."""
#         return {
#             "color": feature["properties"]["default_color"],
#             "weight": feature["properties"]["default_weight"],
#             "opacity": 0.6,
#         }

#     def style_markers(feature):
#         """Basic circleMarker style for plants/demand centers."""
#         color = feature["properties"].get("iconColor", "blue")
#         return {
#             "color": color,
#             "fillColor": color,
#             "fillOpacity": 1.0,
#             "radius": 5,
#         }

#     # -------------------------------------------------
#     # 4) Build Overlays for each technology
#     # -------------------------------------------------
#     tech_overlays = []
#     for tech_name, features in arcs_by_tech.items():
#         if not features:
#             continue
#         geojson_data = {"type": "FeatureCollection", "features": features}
#         overlay = dl.Overlay(
#             dl.GeoJSON(
#                 id=f"{tech_name}-layer",
#                 data=geojson_data,
#                 hoverStyle={"weight": 4, "color": "#666", "dashArray": ""},
#                 options=dict(interactive=True),
#                 # style=style_arcs,
#             ),
#             name=tech_name,
#             checked=True,
#         )
#         tech_overlays.append(overlay)

#     # Single marker overlay for both plants & demand centers
#     markers_overlay = dl.Overlay(
#         dl.GeoJSON(
#             id="markers-layer",
#             data=marker_collection,
#             options=dict(
#                 pointToLayer=assign(
#                     """function(feature, latlng) {
#                 return L.circleMarker(latlng);
#             }"""
#                 )
#             ),
#             # style=style_markers,
#         ),
#         name="Nodes",
#         checked=True,
#     )

#     # -------------------------------------------------
#     # 5) Build the Dash layout
#     # -------------------------------------------------
#     the_map = dl.Map(
#         center=[0, 0],
#         zoom=2,
#         children=[
#             dl.LayersControl(
#                 [dl.BaseLayer(dl.TileLayer(), name="Base", checked=True)] + tech_overlays + [markers_overlay],
#                 id="layers-control",
#             ),
#             dl.ScaleControl(imperial=False, position="bottomleft"),
#         ],
#         style={"width": "100%", "height": "80vh"},
#         id="main-map",
#     )

#     app.layout = html.Div(
#         [
#             html.H3("Dash Leaflet Steel Allocation Map"),
#             html.Button("Refresh Highlight", id="btn-refresh", n_clicks=0),
#             dcc.Store(id="selected-nodes", data=[]),  # holds array of nodeIds
#             the_map,
#         ]
#     )

#     # -------------------------------------------------
#     # 6) Dash callbacks
#     # -------------------------------------------------

#     # (A) Toggle node selection when a user clicks a marker or an arc
#     # In your real code, add more layers for each tech if needed
#     # input_list = [
#     #    Input("markers-layer", "click_feature"),
#     #    Input("BOF-layer", "click_feature"),
#     #    Input("EAF-layer", "click_feature"),
#     #    # Add others if you have them: OHF-layer, ...
#     # ]

#     @app.callback(
#         Output("selected-nodes", "data"),
#         [
#             Input("markers-layer", "click_feature"),
#             Input("BOF-layer", "click_feature"),
#             Input("EAF-layer", "click_feature"),
#             Input("btn-refresh", "n_clicks"),
#         ],
#         State("selected-nodes", "data"),
#     )
#     def toggle_or_refresh(marker_click, bof_click, eaf_click, refresh_click, selected_nodes):
#         ctx = dash.callback_context
#         if not ctx.triggered:
#             return selected_nodes

#         # figure out which input triggered
#         trigger_id = ctx.triggered[0]["prop_id"].split(".")[0]

#         if trigger_id == "btn-refresh":
#             # user clicked the refresh button
#             return []

#         # otherwise, check for arc/marker clicks:
#         if trigger_id == "markers-layer" and marker_click:
#             node_id = marker_click["properties"]["node_id"]
#         elif trigger_id == "BOF-layer" and bof_click:
#             node_id = bof_click["properties"].get("start_node_id") or bof_click["properties"].get("end_node_id")
#         elif trigger_id == "EAF-layer" and eaf_click:
#             node_id = eaf_click["properties"].get("start_node_id") or eaf_click["properties"].get("end_node_id")
#         else:
#             node_id = None

#         if not node_id:
#             return selected_nodes

#         # toggle logic
#         if node_id in selected_nodes:
#             return [nid for nid in selected_nodes if nid != node_id]
#         else:
#             return selected_nodes + [node_id]

#     # (C) Re-style arcs based on selected nodes
#     def highlight_arcs(feature, selected_nodes_list):
#         props = feature["properties"]
#         start_id = props["start_node_id"]
#         end_id = props["end_node_id"]
#         def_color = props["default_color"]
#         def_weight = props["default_weight"]

#         connected = (start_id in selected_nodes_list) or (end_id in selected_nodes_list)
#         if not selected_nodes_list:
#             return {"color": def_color, "weight": def_weight, "opacity": 0.6}
#         elif connected:
#             return {"color": def_color, "weight": def_weight * 3, "opacity": 1.0}
#         else:
#             return {"color": "gray", "weight": def_weight, "opacity": 0.3}

#     def update_geojson_data(original_data, selected_nodes_list):
#         """Return a copy with updated 'style' for each feature."""
#         if not original_data:
#             return {}
#         import copy

#         new_data = copy.deepcopy(original_data)
#         for f in new_data.get("features", []):
#             f["style"] = highlight_arcs(f, selected_nodes_list)
#         return new_data

#     @app.callback(
#         [
#             Output("BOF-layer", "data"),
#             Output("EAF-layer", "data"),
#             # Add more if you have OHF-layer, etc.
#         ],
#         Input("selected-nodes", "data"),
#         [
#             State("BOF-layer", "data"),
#             State("EAF-layer", "data"),
#             # ...
#         ],
#     )
#     def style_arcs_layers(selected_nodes_list, bof_data, eaf_data):
#         bof_out = update_geojson_data(bof_data, selected_nodes_list)
#         eaf_out = update_geojson_data(eaf_data, selected_nodes_list)
#         return (bof_out, eaf_out)
#         # Add more returns if you have more layers.

#     # (D) Optionally hide markers if zoom < 7
#     # @app.callback(
#     #    Output("markers-layer", "data"),
#     #    Input("main-map", "viewport"),
#     #    State("markers-layer", "data"),
#     # )
#     # def hide_markers_on_zoom(viewport, orig_data):
#     #    zoom = viewport.get("zoom", 2)
#     # If we are below zoom 7, return empty data
#     #    if zoom < 7:
#     #        return {"type": "FeatureCollection", "features": []}
#     #    else:
#     #        return orig_data
#     return app



from typing import Self

from ..adapters.repositories import Repository
from ..adapters.repositories.in_memory_repository import InMemoryRepository
from ..domain import Auto


class UnitOfWork:
    repository: Repository

    def __init__(self, repository: Repository = Auto) -> None:
        if not repository:
            repository = InMemoryRepository()
        self.repository = repository
        self.plants = repository.plants
        self.furnace_groups = repository.furnace_groups
        self.demand_centers = repository.demand_centers

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args):
        self.rollback()

    def commit(self):
        pass

    def rollback(self):
        pass

    def collect_new_events(self):
        for plant in self.plants.seen:
            while plant.events:
                yield plant.events.pop(0)



import pandas as pd
from typing import Hashable


def read_and_prep_capex_data(path) -> tuple[pd.DataFrame, pd.Series, pd.Series]:
    """
    Reads the capex data and prepares it for further processing
    """
    # TODO: Might need updating at any point when new data comes in.
    process_design_matrix = pd.read_excel(path, sheet_name="Process step combinations", skiprows=1, index_col=0)
    process_design_matrix.replace("x", 1, inplace=True)
    process_design_matrix.set_index("Process step", inplace=True)
    process_design_matrix.loc["Casting"] = (
        process_design_matrix.loc[pd.Index(["BOF + Casting", "EAF + Casting"])].sum() > 0
    ).astype(int)
    process_design_matrix.index = process_design_matrix.index.str.replace(" + Casting", "")
    capex = pd.read_excel("~/Downloads/Data_collection_ultimate_steel.xlsx", sheet_name="CAPEX", skiprows=2, nrows=20)

    ncapex = capex.set_index("Process step")["Reference unit Capex for 1 Mt plant.1"].dropna()
    # Define the index as a list of strings
    index_values = [
        "Coke production",
        "Sinter production",
        "Pellets production",
        "Shaft Furnace",
        "EAF",
        "Electrolyzer",
        "ESF",
        "BOF",
        "BOF (Relining)",
        "BF",
        "BF (Relining)",
        "Smelting Furnace",
        "Casting",
        "Hot Strip Mill",
        "Oxygen generation",
        "Limestone use",
        "Electricity generation",
        "Steam generation",
        "Other equipment",
    ]
    # Assign the list to the index
    ncapex.index = pd.Index(index_values)
    return process_design_matrix, ncapex, ncapex / 2


def derive_capex_switch(
    process_design_matrix: pd.DataFrame, bf_capex: pd.Series, gf_capex: pd.Series
) -> dict[str | Hashable, dict[str, float]]:
    """
    Takes in dataframes of production process and capex of 1mt installation related to machinery and derives renovation and technology switch capex

    returns a dict that encodes the cost of switch from the technolgy: key to the considered technologies: value
    """
    capex_switch = {}
    for column in process_design_matrix:
        tech_stack = process_design_matrix[column]

        # Use axis parameter instead of "index" string
        # Convert tech_stack.values to Series for sub operation
        difference = process_design_matrix.sub(
            pd.Series(tech_stack.values, index=process_design_matrix.index), fill_value=0
        )
        result = ((difference == 0).mul(bf_capex, axis=0) + (difference == 1).mul(gf_capex, axis=0)).sum().copy()
        # Replace the string safely
        col_name = column
        if isinstance(col_name, str):
            col_name = col_name.replace("Avg ", "")
        capex_switch[col_name] = result.astype(float).round(2).to_dict()
    return capex_switch


def prepare_steel_iron_capex(process_design_matrix: pd.DataFrame, gf_capex: pd.Series) -> dict[str, float]:
    """
    Prepares the full greenfield cost of new furnace
    """

    gf = process_design_matrix.multiply(gf_capex, axis="index").sum()
    gf.index = gf.index.str.replace("Avg ", "")
    return gf.astype(float).round(2).to_dict()


def read_prep_capex_data(path):
    process, ncapex, bf_capex = read_and_prep_capex_data(path)
    full_greenfield = prepare_steel_iron_capex(process, ncapex)
    technology_switch_capex = derive_capex_switch(process, bf_capex, ncapex)
    return full_greenfield, technology_switch_capex


def main(capex_module_path, data_in_path):
    gf_data, switch_data = read_prep_capex_data(data_in_path)

    content = f"""
# Capex swith and full greenfield capex
TECH_SWITCH_CAPEX = {switch_data}
FULL_GREENFIELD_CAPEX = {gf_data}
    """
    with open(capex_module_path, "w") as file:
        file.write(content)


if __name__ == "__main__":
    main("src/steelo/domain/Capex.py", "~/Downloads/Data_collection_ultimate_steel.xlsx")
